///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: computer.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 22/03/04
// Last modification date: 04/03/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 22/03/04	Creation
// 22/10/04	Modify exec() to call writeAllRebuilts()
// 04/03/05	Correct bug in createOutputImage()
//		-> Convert float to double
///////////////////////////////////////////////////////////////////////////////

#include "computer.hh"

using namespace DefectixError;

DefectComputer::DefectComputer(const DefectixOptions::Options& options) : 
  Defectix(options)
{
  if (getError() == NO_ERROR && setError("DefectComputer::DefectComputer", _maskMgr.getError()) == NO_ERROR)
    {
      _evecFile = options.evecFile;
      _nnFile = options.nnFile;
      _nbPC = options.nbPC;
      _PCA = new DefectPCA(options);
      _maskMgr.copyProperties(_imageMgr);
      _outputMgr.copyProperties(_imageMgr);
      _printRebuilt = options.printRebuilt;
      _treshold = options.maskTreshold;
      if (_maskMgr.getError() != NO_ERROR)
	setError("DefectComputer::DefectComputer", _maskMgr.getError());
    }
}

const errorCodes	DefectComputer::exec()
{
  readAllImageExtensions(true);
  if (getError() != NO_ERROR)
    return getError();

  if (loadNetworks() != NO_ERROR)
    return getError();
    
  if (writeAllMasks() != NO_ERROR)
    return getError();

  unreadAllImageExtensions();
  return getError();
}

const errorCodes	DefectComputer::createOutputImage(std::map<int, gsl_matrix_float*>& src,
							  std::map<int, gsl_matrix_float*>& dest,
							  FitsManager& mgr,
							  gsl_matrix_float *evec)
{
  long		nbExt = mgr.getNExtend();
  long		extCnt = 1;
  char		msg[64];
  double	*pc;

  if (src.size() == 0)
    return setError("DefectComputer::createOutputImage", NO_EXT);

  gsl_matrix_float	*inputPart;
  gsl_matrix_float	*outputPart;
  gsl_vector_float	*v = gsl_vector_float_alloc(_blocSize * _blocSize);
  
  pc = new double[_nbPC];
  while (extCnt <= nbExt)
    {
      inputPart = src[extCnt];
      outputPart = dest[extCnt];

      if (inputPart != NULL && outputPart != NULL)
	{
	  for (unsigned row = 0; row < inputPart->size1; ++row)
	    for (unsigned col = 0; col < inputPart->size2; ++col)
	      {
		_PCA->getVectorfromBloc(v, inputPart, row, col);
		gsl_matrix_float_view mv = gsl_matrix_float_view_vector(v, 1, v->size);
		_PCA->computePCA(&(mv.matrix), _nbPC, evec);
		
		gsl_vector_float_const_view pcv = gsl_matrix_float_const_row(_PCA->getPrincipalComponents(), 0);

		for (int cnt1 = 0; cnt1 < _nbPC; ++cnt1)
		  pc[cnt1] = gsl_vector_float_get(&pcv.vector, cnt1);

		feed_forward(_l0, pc);
		if (_treshold == 2)
		  gsl_matrix_float_set(outputPart, row, col, 1 - _l0->output_values[0]);
		else if (_l0->output_values[0] < _treshold)
		  gsl_matrix_float_set(outputPart, row, col, 1);
		else
		  gsl_matrix_float_set(outputPart, row, col, 0);
	      }
	  sprintf(msg, "Mask extension %ld computed", extCnt);
	  printMessage(msg);
	  overwriteMessage();
	}
      else
	{
	  gsl_vector_float_free(v);
	  return setError("DefectComputer::createOutputImage", NO_EXT);
	}
      ++extCnt;
    }
  gsl_vector_float_free(v);
  delete pc;
  return setError("DefectComputer::createOutputImage", NO_ERROR);
}

gsl_matrix_float 	*DefectComputer::loadKLBase()
{
  FILE		*f;

  if ((f = fopen(_evecFile.c_str(), "r")) != NULL)
    {
      unsigned long size1;
      unsigned long size2;
      
      fread(&size1, sizeof(unsigned long), 1, f);
      fread(&size2, sizeof(unsigned long), 1, f);
      gsl_matrix_float *evec = gsl_matrix_float_alloc(size1, size2);
      gsl_matrix_float_fread(f, evec);
      fclose(f);
      printMessage("KL-base loaded");
      return evec;
    }
  else
    {
      setError("DefectComputer::loadKLBase", NO_KLBASE);
      return NULL;
    }
}

const errorCodes	DefectComputer::loadNetworks()
{
  _l0 = load_network_config(_nnFile.c_str());
  printMessage("Network loaded");
  
  gsl_matrix_float *evec = loadKLBase();
  if (getError() != NO_ERROR)
    return getError();

  createOutputImage(_images, _masks, _maskMgr, evec);

  if (_printRebuilt == true)
    writeAllRebuilts(_nbPC);

  free_network(_l0);
  return setError("DefectComputer::loadNetworks", NO_ERROR);
}
