///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: defectix.hh
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 14/01/04
// Last modification date: 04/04/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 14/01/04	Creation using CCfits
// 21/01/04	constructor with default rebin factor = 1
//		readAllImageExtensions() and readImageExtension()
// 26/01/04	Modify readImageExtension() to use CCfits special patch
//		CCfits patch includes
//			readAllKeysNoException()
//			directRead()
// 28/01/04	Doxygen header
//		Error management
//		Bug fix in rebin 
//			if ((j + fac_j < axis[1]) && (i + fac_i < naxis[0]))
// 29/01/04	Bug fix in constructor (memory leak)
// 05/02/04	Move DefectixError to deferror.hh
// 13/02/04	Add PCA methods
// 16/02/04	Bug fix in reallocation of _eval and _evec
// 17/02/04	Normalize input vectors by sqrt(variance)
// 23/02/04	Modify attributes : images and masks are now stored 
//		as gsl_matrix_float
// 26/02/04	Update doxygen header
// 08/03/04	Add getPixel() to get matrix positions corresponding to a image 
//		pixel
//		Add computeBackground() substractBackground() and reduceDyn()
// 10/03/04	Modify attributes to have several FitsManager
//		  one for images and one for masks
//		Add readMaskExtension() and ReadAllMasks()
//		Add virtual exec() for each subclass to be run
// 11/03/04	Update doxygen header
// 15/03/04	Clean code
//		Update error management
//		Move doxygen header to defectix.hh
// 16/03/04	Complete error management *
// 30/06/04	Remove getPixel()
// 22/10/04	Add writeAllRebuilts();
// 04/05/05	create READMODE() macro
///////////////////////////////////////////////////////////////////////////////

#ifndef DEFECTIX_HH_
# define DEFECTIX_HH_


# include <string>
# include <map>
# include <gsl/gsl_sort_vector.h>
# include <gsl/gsl_statistics_float.h>
# include <math.h>
# include "common.hh"

# define READMODE(x) (x == DefectixOptions::TRAINING ? READONLY : READWRITE)

using namespace DefectixError;

class Defectix : public ErrorMgr
{
public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in defectix.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  Defectix(const DefectixOptions::Options& options);
  
  /////////////////////////////////////////////////////////////////////////////
  virtual ~Defectix()
  {
    if (_PCA != NULL)
      delete(_PCA);
  }

  /////////////////////////////////////////////////////////////////////////////
  virtual const errorCodes	exec() = 0;
    
private:
  /////////////////////////////////////////////////////////////////////////////
  // functions in pretreat.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  double	computeBackground(const gsl_vector_float * const pixels, 
				  const unsigned int factor, double stdev);
  
  /////////////////////////////////////////////////////////////////////////////
  void		reduceDyn(gsl_matrix_float *m, s_ext_info *info);
  
  /////////////////////////////////////////////////////////////////////////////
  void		rebin(void **src, long bitpix, const long axes[2]);

  /////////////////////////////////////////////////////////////////////////////
  void		computeInfo(gsl_matrix_float *m, gsl_vector_float *pixels, s_ext_info *info);

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  void		rebin(T **src, T *dest, const long axes[2]);

protected:
  /////////////////////////////////////////////////////////////////////////////
  void		substractBackground(gsl_matrix_float *m, s_ext_info *info);

public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in read.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  errorCodes	readImageExtension(int pos, bool createMask = false);
  
  /////////////////////////////////////////////////////////////////////////////
  errorCodes	readAllImageExtensions(bool createMask = false);

  /////////////////////////////////////////////////////////////////////////////
  errorCodes	Defectix::reduceAllImagesDyn();

  /////////////////////////////////////////////////////////////////////////////
  void		unreadImageExtension(int pos);

  /////////////////////////////////////////////////////////////////////////////
  void		unreadAllImageExtensions();

  /////////////////////////////////////////////////////////////////////////////
  errorCodes	readMaskExtension(int pos);

  /////////////////////////////////////////////////////////////////////////////
  errorCodes	readAllMasks();
  
  /////////////////////////////////////////////////////////////////////////////
  void		unreadMaskExtension(int pos);

  /////////////////////////////////////////////////////////////////////////////
  void		unreadAllMasks();


protected:
  /////////////////////////////////////////////////////////////////////////////
  // function in write.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  const errorCodes	writeAllMasks();

  /////////////////////////////////////////////////////////////////////////////
  const errorCodes	writeAllRebuilts(int nbPC);

private:
  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  errorCodes	commonRead(gsl_matrix_float *m, T **src, FitsManager& mgr);

  /////////////////////////////////////////////////////////////////////////////
  errorCodes	commonRead(int pos, gsl_matrix_float *m, FitsManager& mgr);
  
  /////////////////////////////////////////////////////////////////////////////
  void		commonPrint(int pos, const char * const type) const;

protected:
  /////////////////////////////////////////////////////////////////////////////
  // message management							     //
  /////////////////////////////////////////////////////////////////////////////
  inline void		printMessage(std::string msg) const
  {
    std::cout << MESSAGE_START << msg << std::endl;    
  }
  
  ///////////////////////////////////////////////////////////////////////////
  inline void		overwriteMessage() const
  {
    std::cout << "\033[1A\033[1M";
  }

protected:
  ////////////////
  // attributes //
  ////////////////
  FitsManager			_imageMgr;
  FitsManager			_maskMgr;
  FitsManager			_outputMgr;
  int				_factor;
  int				_blocSize;
  int				_zoomStep;
  bool				_substractBg;
  bool				_reduceDyn;
  std::map<int, gsl_matrix_float*>	_images;
  std::map<int, t_ext_info>	_imagesInfo;
  std::map<int, gsl_matrix_float*>	_masks;
  DefectPCA			*_PCA;
};

#endif /* !DEFECTIX_HH_ */
