///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: deferror.hh
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 05/02/04
// Last modification date: 15/03/04
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 05/02/04	Copy namespace defectixError from defectix.hh
// 15/03/04	Add error codes NOT_ALLOC_MAT, NOT_ALLOC_TABLE 
//		Move doxygen header to deferror.dox
// 16/03/04	Create Error class to manage errors in any subclass
///////////////////////////////////////////////////////////////////////////////

#ifndef DEFECTIX_ERROR_HH_
# define DEFECTIX_ERROR_HH_

# include <string>
# include <iostream>
# include <fitsio.h>
# include "defoptions.hh"

namespace DefectixError
{
  enum			errorCodes {NO_ERROR, NO_EXTENSION, MALLOC, FITS, 
				    NOT_MEF, NOT_IMGEXT, NO_EXT, NO_MASK, 
				    WRONG_BITPIX, MASK_SIZE,
				    NOT_ALLOC_MAT, NOT_ALLOC_TABLE,
				    NO_KLBASE,
				    CREATE_FILE};

  struct	sErrorMsgs
  {
    int		error;
    char	*msg;
  };
  
  static struct sErrorMsgs errorMsgs[] =
    {
      {NO_EXTENSION, "try to use an extension not loaded"},
      {MALLOC, "cannot allocate memory"},
      {FITS, "cfitsio library error returned"},
      {NOT_MEF, "file is not a standard MEF"},
      {NOT_IMGEXT, "extension is not an image"},
      {NO_EXT, "no extension at given position"},
      {NO_MASK, "no mask at given position"},
      {WRONG_BITPIX, "bitpix not supported"},
      {MASK_SIZE, "mask dimensions are different from image ones"},
      {NOT_ALLOC_MAT, "matrix not allocated before use"},
      {NOT_ALLOC_TABLE, "table not allocated before use"},
      {NO_KLBASE, "cannot load KL-base"},
      {0, NULL}
    };

  class ErrorMgr
  {
  public:
    ErrorMgr(const bool verbose) : 
      _error(NO_ERROR), _verbose(verbose)
    {
    }

  protected:
    ///////////////////////////////////////////////////////////////////////////
    // error management							     //
    ///////////////////////////////////////////////////////////////////////////
    inline const errorCodes	setError(std::string c, errorCodes e)
    {
      if (_verbose)
	std::cout << c << ": set error " << e << std::endl;

      if (_error == NO_ERROR)
	_error = e;
      return _error;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    inline void		setFitsError(int status)
    {
      if (status)
	{
	  setError("ErrorMgr::setFitsError", FITS);
	  fits_report_error(stderr, status);
	}
    }

  public:
    ///////////////////////////////////////////////////////////////////////////
    inline const errorCodes	getError() const
    {
      return _error;
    }
    
    inline void		resetError()
    {
      if (_verbose)
	std::cout << "reset error " << _error << std::endl;
      _error = NO_ERROR;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    const errorCodes	printError() const
    {
      if (_error == NO_ERROR)
	{
	  std::cerr << MESSAGE_START << "no error (0)" << std::endl;
	  return _error;
	}
      
      for (int i = 0; errorMsgs[i].msg != NULL; i++)
	{
	  if (_error == errorMsgs[i].error)
	    {
	      std::cerr << ERROR_START << errorMsgs[i].msg 
			<< " (" << _error << ")" << std::endl;
	      return _error;
	    }
	}
      std::cerr << ERROR_START << "unknown error (" 
		<< _error << ")" << std::endl;
      return _error;
    }
  
  protected:
    errorCodes	_error;
    bool	_verbose;
  };
}

#endif /* !DEFECTIX_ERROR_HH_ */
