///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: defectix.hh
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 24/02/04
// Last modification date: 26/02/04
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 24/02/04	Create namespace DefectixOptions
// 26/02/04	Doxygen header
// 10/03/04	Add options for execution mode
//		Add routines to manage configuration files
// 11/03/04	Update doxygen header
// 15/03/04	Mode doxygen header to defoptions.dox
///////////////////////////////////////////////////////////////////////////////

#ifndef DEFECTIX_OPTIONS_HH_
# define DEFECTIX_OPTIONS_HH_

# include "getopt.h"
# include "config.h"

namespace DefectixOptions
{
  enum		mode{HALO, TRAIL, TRAINING, COMPUTING, NOMODE};
  enum		simulatortype{EXTENSION, MOSAIC, NOTYPE};
  
  template <typename T>
  struct	sOptions
  {
    T		value;
    char	*name;
  };
  
  class		Options
  {
  public:
    Options() : mode(DefectixOptions::NOMODE),
		//
		confFile(""), inputFile(""), maskFile("!mask.fits"), outputFile("!res.fits"),
		evecFile(""), nnFile(""),
		rebinFactor(4), substractBg(true), reduceDyn(true),
		//
		nbDefects(5), simulatorType(DefectixOptions::EXTENSION), 
		intensity(200), nbVectors(500000),
		//
		blocSize(8), nbPC(32), zoomStep(5),
		printEvec(false), printRebuilt(false),
		//
		energyTreshold(0.025), maskTreshold(2.0), maxIter(200),
		maxPrior(1.0), minPrior(0.0), nonPrior(1.0),
		//
		verbose(false)
    {
    }

    void reinit()
    {
      confFile.erase(); 
      inputFile.erase();
      evecFile.erase();
      maskFile = "!mask.fits";
      outputFile = "!res.fits";

      rebinFactor = 4; 
      substractBg = true; 
      reduceDyn = true; 

      nbDefects = 5;
      simulatorType = DefectixOptions::EXTENSION;
      intensity = 200;
      nbVectors = 500000;
      
      blocSize = 8; 
      nbPC = 32; 
      zoomStep = 5; 

      printEvec = false; 
      printRebuilt = false;

      energyTreshold = 0.025;
      maskTreshold = 2.0;
      maxIter = 200;

      maxPrior = nonPrior = 1.0;
      minPrior = 0.0;

      verbose = false;
    }
    
  public:
    // exec mode
    DefectixOptions::mode	mode;
    // files
    std::string			confFile;
    std::string			inputFile;
    std::string			maskFile;
    std::string			outputFile;
    std::string			evecFile;
    std::string			nnFile;
    // loading options
    int				rebinFactor;
    bool			substractBg;
    bool			reduceDyn;	

    // defadder options
    int				nbDefects;
    DefectixOptions::simulatortype	simulatorType;
    int				intensity;
    int				nbVectors;

    // PCA options
    int				blocSize;
    int				nbPC;
    int				zoomStep;
    // PCA outputs
    bool			printEvec;
    bool			printRebuilt;
    // Neural network
    double			energyTreshold;
    double			maskTreshold;
    unsigned			maxIter;
    // Priors
    double			maxPrior;
    double			minPrior;
    double			nonPrior;
    // verbose mode
    bool			verbose;
  };
  
  enum options_e
    {
      opt_help = 'h',
      opt_version = 'V',
      opt_default = 'd',
      opt_mode = 'M',
      opt_conf = 'C',
      opt_input = 'I',
      opt_mask = 'K',
      opt_output = 'O',
      opt_efile = 'L',
      opt_nnfile = 'N',
      opt_rebin = 'r',
      opt_bsize = 'b',
      opt_nbpc = 'p',
      opt_zoom = 'z',
      opt_evec = 'E',
      opt_rebuilt = 'R',
      opt_subbg = 'S',
      opt_dyn = 'D',
      opt_intensity = 'i',
      opt_addertype = 'a',
      opt_nbdef = 'f',
      opt_verbose = 'v',
      opt_nbvec = 't',
      opt_maxiter = 'm',
      opt_energy = 'e',
      opt_mtreshold = 's',
      opt_maxprior = 'x',
      opt_minprior = 'n',
      opt_nonprior = 'l',
    };
  
  
  void		usage(void);
  void		version(void);
  void		defaultconf(int mode);
  void		set_option(int c, Options& options);
  bool		set_file_option(char *line, char *keyword, 
				int key_opt, Options &options);
  int		conf_file_error(int line);
  void		read_conf_file(FILE *f, Options& options);
}

#endif /* !DEFECTIX_OPTIONS_HH_ */
