///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: fitsmgr.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 18/03/04
// Last modification date: 15/05/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 18/03/04	Move code from fitsmgr.hh
//		Contain general and fetch methods
//			constructor and destructor
//			copyProperties()
//			pHDU()
//			extHDU()
// 10/01/05	Add code to manage single FITS files
//			uniqueHDU()
//		Modify constructor to manage single FITS files
// 22/03/05	Modify constructor to manage single FITS files 
//		according to EXTEND flag and not only NEXTEND value
// 06/04/05	Modify constructor to handle "EXTEND=T" according to
//		FITS fucking format
// 02/05/05	Debug constructor modified on 06/04/05
// 13/05/05	Add copyHDU()
///////////////////////////////////////////////////////////////////////////////

# include "fitsmgr.hh"

///////////////////////////////////////////////////////////////////////////////
// unique constructor							     //
///////////////////////////////////////////////////////////////////////////////
FitsManager::FitsManager(const std::string& fileName, int mode, bool verbose) :
  ErrorMgr(verbose)
{
  unsigned short	value;
  int			status(0);
  struct stat		stat;

  _error = DefectixError::NO_ERROR;
  _name = fileName;
  _maxdim = 0;
  _currentExt = -1;
  _naxes = _fpixel = NULL;
  _file = NULL;

  if (mode == READWRITE)
    fits_create_file(&_file, _name.c_str(), &status);
  else if (lstat(_name.c_str(), &stat) == 0)
    {
      fits_open_file(&_file, _name.c_str(), mode, &status);
      if (status)
	setFitsError(status);
      fits_read_key(_file, TUSHORT, "SIMPLE", &value, NULL, &status);
      fits_read_key(_file, TINT, "BITPIX", &_bitpix, NULL, &status);

      if (status)
	setFitsError(status);
      else
	{
	  fits_read_key(_file, TINT, "EXTEND", &value, NULL, &status);

	  // "EXTEND" keyword unavailable
	  if (status)
	    {
	      _type = NO_EXTENDED;
	      _nextend = 1;
	      status = 0;
	    }
	  // "EXTEND" keyword available
	  else
	    {
	      _type = EXTENDED;
	      fits_read_key(_file, TINT, "NEXTEND", &_nextend, NULL, &status);

	      // NEXTEND keyword unavailable
	      if (status)
		{
		  fits_movabs_hdu(_file, 2, &_currentType, &status);
		  // No extensions
		  if (status)
		    {
		      _type = NO_EXTENDED;
		      _nextend = 1;
		      status = 0;
		    }
		  // Compute number of extensions
		  else
		    {
		      _nextend = 1;
		      while (!status)
			{
			  fits_movabs_hdu(_file, _nextend + 2, &_currentType, &status);
			  _nextend++;
			}
		      status = 0;
		    }
		}
	    }
	}
    }
  else
    setFitsError(104);

  if (status)
    {
      fits_report_error(stderr, status);
      setFitsError(status);
    }
}

FitsManager::~FitsManager()
{
  int	status;
  
  delete (_naxes);
  delete (_fpixel);
  fits_close_file(_file, &status);
}

void	FitsManager::copyProperties(FitsManager& src)
{
  int	status(0);
  
  _error = DefectixError::NO_ERROR;
  _maxdim = src._maxdim;
  _currentExt = src._currentExt;
  _currentType = src._currentType;
  _naxis = src._naxis;
  _naxes = new long(_maxdim);
  _fpixel = new long(_maxdim);
  _file = NULL;
  _bitpix = src._bitpix;
  _nextend = src._nextend;
  _type = src._type;
  
  fits_create_file(&_file, _name.c_str(), &status);
  if (_type == EXTENDED)
    {
      src.pHDU();
      fits_copy_hdu(src._file, _file, 0, &status);
    }

  if (status)
    {
      fits_report_error(stderr, status);
      setError("FitsManager::copyProperties", DefectixError::FITS);
    }
}

///////////////////
// fetch methods //
///////////////////
void	FitsManager::pHDU()
{
  int	status;

  status = 0;
  fits_movabs_hdu(_file, 1, &_currentType, &status);
  setFitsError(status);
}

void	FitsManager::extHDU(int pos)
{
  int	status;
  int	tmp;

  if (pos > 0 && pos <= _nextend)
    {
      status = 0;
      _currentExt = pos;
      fits_movabs_hdu(_file, pos + 1, &_currentType, &status);
      if (_currentType != IMAGE_HDU)
	setError("FitsManager::extHDU", DefectixError::NOT_IMGEXT);
      else
	{
	  fits_read_key(_file, TINT, "NAXIS", &tmp, NULL, &status);
	  if (tmp != _maxdim)
	    {
	      _maxdim = tmp;
	      delete (_naxes);
	      delete (_fpixel);
	      _naxes = new long(_maxdim);
	      _fpixel = new long(_maxdim);
	      for (tmp = 0; tmp < _maxdim; tmp++)
		_fpixel[tmp] = 1;
	    }
	  fits_get_img_param(_file, _maxdim, &_bitpix, &_naxis, _naxes, &status);
	  setFitsError(status);
	}
    }
  else
    setError("FitsManager::extHDU", DefectixError::NO_EXT);
}

void	FitsManager::uniqueHDU()
{
  int	status;
  int	tmp;

  status = 0;
  fits_read_key(_file, TINT, "NAXIS", &tmp, NULL, &status);
  if (tmp != _maxdim)
    {
      _maxdim = tmp;
      delete (_naxes);
      delete (_fpixel);
      _naxes = new long(_maxdim);
      _fpixel = new long(_maxdim);
      for (tmp = 0; tmp < _maxdim; tmp++)
	_fpixel[tmp] = 1;
    }
  fits_get_img_param(_file, _maxdim, &_bitpix, &_naxis, _naxes, &status);
  setFitsError(status);
}

void	FitsManager::copyHDU(FitsManager& src)
{
  int	status;

  fits_copy_hdu(src._file, _file, 0, &status);
  std::cout << "cpy" << std::endl;
  if (status)
    {
      fits_report_error(stderr, status);
      setError("FitsManager::copyProperties", DefectixError::FITS);
    }
}
