///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: fitsmgr.hh
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 05/02/04
// Last modification date: 13/05/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 05/02/04	Creation of a class to simple fits file management
// 11/02/04	Doxygen header
// 26/02/04	Add getAxes() and update doxygen header
// 10/03/04	Modify constructor according to r/w mode
//		Add copyProperties()
//		Add setHDUAsImage() and writeImage()
// 15/03/04	Move doxygen header to fitsmgr.dox
//		Clean code
//		Complete error management *
// 18/03/04	Move code to fitsmgr.cc
// 10/01/05	Add code to manage single FITS files
// 13/05/05	Add copyHDU()
///////////////////////////////////////////////////////////////////////////////

#ifndef FITS_MANAGER_HH_
# define  FITS_MANAGER_HH_

# include <fitsio.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <unistd.h>
# include "common.hh"

typedef struct	s_ext_info
{
  double	min;
  double	max;
  double	bg;
}		t_ext_info;

class FitsManager : public DefectixError::ErrorMgr
{
public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in fitsmgr.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  FitsManager(const std::string& fileName, int mode, bool verbose);

  /////////////////////////////////////////////////////////////////////////////
  ~FitsManager();

  /////////////////////////////////////////////////////////////////////////////
  void	copyProperties(FitsManager& src);


public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in fitsmgr.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  void	pHDU();

  /////////////////////////////////////////////////////////////////////////////
  void	extHDU(int pos);

  /////////////////////////////////////////////////////////////////////////////
  void	uniqueHDU();

  /////////////////////////////////////////////////////////////////////////////
  void	copyHDU(FitsManager& src);

public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in fitsread.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  void	readImage(void **img);
  
  /////////////////////////////////////////////////////////////////////////////
  void	unreadImage(void **img);

public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in fitswrite.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  void	setHDUAsImage(int bitpix, 
		      const long naxis, 
		      const long *axes, 
		      const long factor);
  
  /////////////////////////////////////////////////////////////////////////////
  void	writeImage(void **img);
    
public:
  /////////////////////////////////////////////////////////////////////////////
  // accessors								     //
  /////////////////////////////////////////////////////////////////////////////
  inline const int	getBitpix() const
  {
    return _bitpix;
  }

  inline const long	getAxis(int i) const
  {
    if (i >= 0 && i < _naxis)
      return _naxes[i];
    else
      return 0;
  }

  inline const long	*getAxes() const
  {
    return _naxes;
  }

  inline const long	getNAxes() const
  {
    return _naxis;
  }

  inline const long	getNExtend() const
  {
    return _nextend;
  }

  inline const int	getType() const
  {
    return _type;
  }

private:
  /////////////////////////////////////////////////////////////////////////////
  // attributes								     //
  /////////////////////////////////////////////////////////////////////////////
  // pointer to the cfitsio structure
  fitsfile	*_file;

  // variables read in the primary HDU
  std::string	_name;
  int		_nextend;
  int		_bitpix;

  // internal data about current extension
  int		_currentExt;
  int		_currentType;
  
  // variables read for each extension
  int		_maxdim;
  int		_naxis;
  long		*_naxes;
  long		*_fpixel;

  // type of the fits file
  // EXT_FILE : A primary header and image extensions
  // NO_EXT_FILE : Primary header IS the image
  int		_type;

public:
  enum	e_fits_type{EXTENDED, NO_EXTENDED};
};

#endif /* !FITS_MANAGER_HH_ */
