/******************************************************************************
** Project: neural_network
** Filename: batch_back_prop.c
*******************************************************************************
** Authors: A. BAILLARD & T. BAUDLOT
** Creation date: 14/07/03
** Last modification date: 14/07/03
*******************************************************************************
** Comments:
** 14/07/03	AB	Functions for batch back propagation
**				- init_batch_iteration()
**				- batch_back_propagation()
** 16/07/03	TB	Creation of an output file that holds energy for 
**			each iteration
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "network.h"

/**
 ** create_deltas() allocates vectors for variations
 ** parameters:
 **	network: network being trained
 **     delta_hidden_weights: vector of hidden weights variations
 **     delta_output_weights: vector of output weights variations
 **     delta_hidden_biases: vector of hidden biases variations
 **     delta_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void	create_deltas(t_network *network,
			      double	***delta_hidden_weights,
			      double	***delta_output_weights,
			      double	**delta_hidden_biases,
			      double	**delta_output_biases)
{
  unsigned i;
  
  *delta_output_biases = (double *)xmalloc(network->nb_output_units * 
						 sizeof(double));

  *delta_hidden_biases = (double *)xmalloc(network->nb_hidden_units * 
						 sizeof(double));


  *delta_output_weights = (double **)xmalloc(network->nb_output_units * 
					    sizeof(double *));
  for (i = 0; i < network->nb_output_units; i++)
    (*delta_output_weights)[i] = (double *)xmalloc(network->nb_hidden_units * 
						 sizeof(double));
  
  *delta_hidden_weights = (double **)xmalloc(network->nb_hidden_units * 
					     sizeof(double *));
  for (i = 0; i < network->nb_hidden_units; i++)
    (*delta_hidden_weights)[i] = (double *)xmalloc(network->nb_input_units * 
						 sizeof(double));
}


/**
 ** delete_deltas() frees vectors for variations
 ** parameters:
 **	network: network being trained
 **     delta_hidden_weights: vector of hidden weights variations
 **     delta_output_weights: vector of output weights variations
 **     delta_hidden_biases: vector of hidden biases variations
 **     delta_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void	delete_deltas(t_network *network,
			      double	**delta_hidden_weights,
			      double	**delta_output_weights,
			      double	*delta_hidden_biases,
			      double	*delta_output_biases)
{
  unsigned	i;
  for (i = 0; i < network->nb_output_units; i++)
    free(delta_output_weights[i]);
  
  for (i = 0; i < network->nb_hidden_units; i++)
    free(delta_hidden_weights[i]);
  
  free(delta_output_weights);
  free(delta_hidden_weights);
  free(delta_hidden_biases);
  free(delta_output_biases);
}


/**
 ** init_deltas() inits weights and biases to zero
 ** parameters:
 **	network: network being trained
 **     delta_hidden_weights: vector of hidden weights variations
 **     delta_output_weights: vector of output weights variations
 **     delta_hidden_biases: vector of hidden biases variations
 **     delta_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void	init_deltas(t_network *network, 
			    double	**delta_hidden_weights,
			    double	**delta_output_weights,
			    double	*delta_hidden_biases,
			    double	*delta_output_biases)
{
  unsigned	i;
  unsigned	j;

  for (i = 0; i < network->nb_output_units; i++)
    {
      delta_output_biases[i] = 0;
      for (j = 0; j < network->nb_hidden_units; j++)
	delta_output_weights[i][j] = 0;
    }

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      delta_hidden_biases[i] = 0;
      for (j = 0; j < network->nb_input_units; j++)
	delta_hidden_weights[i][j] = 0;
    }
}


/**
 ** compute_deltas() computes variations of weights and biases
 ** parameters:
 **	network: network being trained
 **     delta_hidden_weights: vector of hidden weights variations
 **     delta_output_weights: vector of output weights variations
 **     delta_hidden_biases: vector of hidden biases variations
 **     delta_output_biases: vector of output biases variations
 **	input_sample: sample currently used for training
 ** return value:
 **	none
 **/
static void	compute_deltas(t_network *network, 
			       double	**delta_hidden_weights,
			       double	**delta_output_weights,
			       double	*delta_hidden_biases,
			       double	*delta_output_biases,
			       double	*input_sample)
{
  unsigned	i;
  unsigned	j;

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	{
	  delta_hidden_weights[i][j] += 
	    network->hidden_energy[i] * input_sample[j];
	}
      delta_hidden_biases[i] += network->hidden_energy[i];
    }

  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	{
	  delta_output_weights[i][j] +=
	    network->output_energy[i] * network->hidden_values[j];
	}
      delta_output_biases[i] += network->output_energy[i];      
    }
}


/**
 ** update_network() computes new weights and biases
 ** parameters:
 **	network: network being trained
 **     delta_hidden_weights: vector of hidden weights variations
 **     delta_output_weights: vector of output weights variations
 **     delta_hidden_biases: vector of hidden biases variations
 **     delta_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void	update_network(t_network *network, 
			       double	**delta_hidden_weights,
			       double	**delta_output_weights,
			       double	*delta_hidden_biases,
			       double	*delta_output_biases)
{
  unsigned	i;
  unsigned	j;

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	{
	  network->hidden_weights[i][j] -= 
	    network->epsilon * delta_hidden_weights[i][j];
	}
      network->hidden_biases[i] -=
	network->epsilon * delta_hidden_biases[i];
    }
  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	{
	  network->output_weights[i][j] -= network->epsilon * 
	    delta_output_weights[i][j];
	}
      network->output_biases[i] -=
	network->epsilon * delta_output_biases[i];
    }
}


/**
 ** batch_back_propagation() trains the network on a set of examples 
 ** composed of an input and an expected output
 ** parameters:
 **	network: allocated network to train
 **	base_dim: dimension of the examples base
 **	input_base: set of input vectors
 **	n_input: size of an input vector
 **	expected_base: set of output vectors corresponding to input vectors
 **	n_output: size of an expected output vector
 ** return value:
 **	error code
 **/
int		batch_back_propagation(t_network *network, unsigned base_dim,
				       double **input_base, unsigned n_input,
				       double **expected_base, unsigned n_output)
{
  unsigned	sample_cnt;
  double	**delta_hidden_weights;
  double	**delta_output_weights;
  double	*delta_hidden_biases;
  double	*delta_output_biases;
  double	energy;
  double	previous_energy;
  unsigned	iter;
  unsigned	nb_iter;
  FILE		*file;

  create_deltas(network, &delta_hidden_weights, &delta_output_weights,
		&delta_hidden_biases, &delta_output_biases);

  energy = 0;
  iter = 0;
  nb_iter = 0;
  file = fopen("batch_backprop.res","w");
  do
    {
      previous_energy = energy;
      energy = 0;
      init_deltas(network, delta_hidden_weights, delta_output_weights,
			   delta_hidden_biases, delta_output_biases);

      for (sample_cnt = 0; sample_cnt < base_dim; sample_cnt++)
	{
	  feed_forward(network, input_base[sample_cnt]);
	  
	  compute_gradients(network, expected_base[sample_cnt]);
	  
	  compute_deltas(network, delta_hidden_weights, delta_output_weights,
			 delta_hidden_biases, delta_output_biases, input_base[sample_cnt]);
	  
	  compute_energy(network, expected_base[sample_cnt]);
	  energy += network->energy;
	}
      energy /= base_dim;

      update_network(network, delta_hidden_weights, delta_output_weights,
		     delta_hidden_biases, delta_output_biases);
      
      
      fprintf(file, "%d\t%f\n", iter, energy);
      ++iter;
      if ((energy - previous_energy) < 0.000001 && (energy - previous_energy) > -0.000001)
	++nb_iter;
      else
	nb_iter = 0;
    }
  while (energy > network->energy_treshold && nb_iter < network->max_iter);

  fclose(file);

  delete_deltas(network, delta_hidden_weights, delta_output_weights,
		delta_hidden_biases, delta_output_biases);

  return LIBNN_NO_ERROR;
}
