/******************************************************************************
** Project: neural_network
** Filename: network.c
*******************************************************************************
** Authors: A. BAILLARD & T. BAUDLOT
** Creation date: 01/07/03
** Last modification date: 15/04/04
*******************************************************************************
** Comments:
** 01/07/03	AB	Main functions to handle network
**				- create_network()
**				- free_network()
**				- print_network()
** 04/07/03	AB	Add train functions
**				- back_propagation()
**				- compute_energy()
**				- feed_forward()
** 08/07/03	TB	Random Weights initialisation with tiny values 
** 11/07/03	TB	Add biases for output and hidden layers
** 14/07/03	AB	Train functions moved to specific files
**				prop.c
**				stoch_back_prop.c
**				batch_back_prop.c
** 15/04/04	AB	Add an activation function (SIGMOID_0_1)
******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include "network.h"
#include "sig.h"

/**
 ** Global variable for signal handling
 **/
short	sig_catched = 0;

/**
 ** List of available activation functions
 ** Functions coded in activation_function.c
 **/
t_acfunc	activation_functions[] = 
{
  {"SIGMOID", SIGMOID, sigmoid, sigmoid_drv, 1},
  {"SIGMOID_0_1", SIGMOID_0_1, sigmoid_0_1, sigmoid_0_1_drv, 1},
  {NULL, END, NULL, NULL, 0}
};


/**
 ** create_network() returns a allocated network with the indicated number 
 ** of neurons and the chosen activation function
 ** parameters:
 **	nb_input_nodes: number of units in the input layer
 **	nb_hidden_nodes: number of units in the hidden layer
 **	nb_input_nodes: number of units in the output layer
 **	acfunc: string corresponding to an available activation function
 ** return value
 **	allocated network
 **/
t_network	*create_network(unsigned nb_input_nodes, 
				unsigned nb_hidden_nodes, 
				unsigned nb_output_nodes, 
				char	 *acfunc)
{
  unsigned     	cnt;
  char		*upper_acfunc;

  upper_acfunc = toupper_string(acfunc);

  for (cnt = 0; activation_functions[cnt].id != END; cnt++)
    if (!strcmp(activation_functions[cnt].name, upper_acfunc))
      {
	return create_network_common(nb_input_nodes, nb_hidden_nodes, 
				     nb_output_nodes, activation_functions[cnt]);
      }
	
  printf("Unknown activation function %s\n", acfunc);
  return NULL;
}

/**
 ** create_network() returns a allocated network with the indicated number 
 ** of neurons and the chosen activation function
 ** parameters:
 **	nb_input_nodes: number of units in the input layer
 **	nb_hidden_nodes: number of units in the hidden layer
 **	nb_input_nodes: number of units in the output layer
 **	acfunc: enum value corresponding to an available activation function
 ** return value
 **	allocated network
 **/
t_network	*create_network2(unsigned nb_input_nodes, 
				unsigned nb_hidden_nodes, 
				unsigned nb_output_nodes, 
				int	 acfunc)
{
  unsigned     	cnt;

  for (cnt = 0; activation_functions[cnt].id != END; cnt++)
    if (activation_functions[cnt].id == acfunc)
      {
	return create_network_common(nb_input_nodes, nb_hidden_nodes, 
				     nb_output_nodes, activation_functions[cnt]);
      }
	
  printf("Unknown activation function %d\n", acfunc);
  return NULL;
}

/**
 ** create_network2() returns a allocated network with the indicated number 
 ** of neurons and the chosen activation function
 ** parameters:
 **	nb_input_nodes: number of units in the input layer
 **	nb_hidden_nodes: number of units in the hidden layer
 **	nb_input_nodes: number of units in the output layer
 **	acfunc: structure corresponding to an available activation function
 ** return value
 **	allocated network
 **/
t_network	*create_network_common(unsigned nb_input_nodes, 
				unsigned nb_hidden_nodes, 
				unsigned nb_output_nodes, 
				t_acfunc acfunc)
{
  unsigned     	cnt;
  t_network	*network;

  network = (t_network *)xmalloc(sizeof(t_network));
  
  network->nb_input_units = nb_input_nodes;
  network->nb_hidden_units = nb_hidden_nodes;
  network->nb_output_units = nb_output_nodes;

  network->acfunc = acfunc;

  network->hidden_activation = (double *)xmalloc(nb_hidden_nodes * 
						 sizeof(double));
  network->hidden_values = (double *)xmalloc(nb_hidden_nodes * 
						 sizeof(double));
  network->hidden_biases = (double *)xmalloc(nb_hidden_nodes * 
						 sizeof(double));

  network->output_activation = (double *)xmalloc(nb_output_nodes * 
						 sizeof(double));
  network->output_values = (double *)xmalloc(nb_output_nodes * 
						 sizeof(double));
  network->output_biases = (double *)xmalloc(nb_output_nodes * 
						 sizeof(double));

  network->hidden_weights = (double **)xmalloc(nb_hidden_nodes * 
					       sizeof(double *));
  for (cnt = 0; cnt < nb_hidden_nodes; cnt++)
    network->hidden_weights[cnt] = (double *)xmalloc(nb_input_nodes * 
						     sizeof(double));

  network->output_weights = (double **)xmalloc(nb_output_nodes * 
					       sizeof(double *));
  for (cnt = 0; cnt < nb_output_nodes; cnt++)
    network->output_weights[cnt] = (double *)xmalloc(nb_hidden_nodes * 
						     sizeof(double));

  network->hidden_energy = (double *)xmalloc(nb_hidden_nodes * 
						 sizeof(double));

  network->output_energy = (double *)xmalloc(nb_output_nodes * 
						 sizeof(double));

  network->max_iter = MAX_ITER;
  network->energy_treshold = ENERGY_TRESHOLD;

  signal(SIGINT, my_signal);
  return network;
}

/**
 ** init_network() initialize values
 ** randomize weights beetwen -0.1 and 0.1
 ** initialize biases to 1
 **/
void		init_network(t_network *network)
{
  unsigned     	i;
  unsigned     	j;

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	network->hidden_weights[i][j] = (random() % 1000) / 10000.0 - 0.05;
      network->hidden_biases[i] = 1;
    }

  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	network->output_weights[i][j] = (random() % 1000) / 10000.0 - 0.05;
      network->output_biases[i] = 1;
    }
  network->epsilon = EPSILON;
}


/**
 ** feed_forward() gives an input vector to the network
 ** network computes corresponding output vector
 ** size of the input vector must be checked before the call
 ** parameters:
 **	network: allocated network to feed
 **	input: input vector
 ** return value:
 **	none
 **/
void		feed_forward(t_network *network, double *input)
{
  unsigned     	i;
  unsigned     	j;

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      network->hidden_activation[i] = network->hidden_biases[i];
      for (j = 0; j < network->nb_input_units; j++)
	network->hidden_activation[i] += network->hidden_weights[i][j] * input[j];
      network->hidden_values[i] = 
	network->acfunc.func(network->hidden_activation[i], network->acfunc.lambda);
    }

  for (i = 0; i < network->nb_output_units; i++)
    {
      network->output_activation[i] = network->output_biases[i];
      for (j = 0; j < network->nb_hidden_units; j++)
	network->output_activation[i] +=  network->output_weights[i][j] * 
	  network->hidden_values[j];
      network->output_values[i] = 
	network->acfunc.func(network->output_activation[i], network->acfunc.lambda);
    }
}


/**
 ** compute_energy() returns energy corresponding to the last exemple learned
 ** expected_output must be the last vector that fed the network
 ** parameters:
 **	network: allocated network to consider
 **	expected_output: output vector to compare with network results
 ** return value:
 **	corresponding energy
 **/
double  	compute_energy(t_network *network, double *expected_output)
{
  unsigned     	i;

  network->energy = 0;

  for (i = 0; i < network->nb_output_units; i++)
    network->energy += (network->output_values[i] - expected_output[i]) * 
      (network->output_values[i] - expected_output[i]);
  return network->energy;
}
     
/**
 ** free_network() frees a previously allocated network
 ** parameters:
 **	network: allocated network to free
 ** return value:
 **	none
 **/
void		free_network(t_network *network)
{
  unsigned     	cnt;
  
  for (cnt = 0; cnt < network->nb_hidden_units; cnt++)
    free(network->hidden_weights[cnt]);
  free(network->hidden_weights);

  for (cnt = 0; cnt < network->nb_output_units; cnt++)
    free(network->output_weights[cnt]);
  free(network->output_weights);

  free(network);
}


/**
 ** print_network() prints weight values for hidden and output layers
 ** parameters:
 **	network: allocated network to print on stdout
 ** return value:
 **	none
 **/
void		print_network(t_network *network)
{
  unsigned     	i;
  unsigned     	j;

  printf("*********** Network ************\n");
  printf("Hidden weights\n");

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      printf("[{");
      for (j = 0; j < network->nb_input_units; j++)
	printf("%f ", network->hidden_weights[i][j]);
      printf("}, {%f}]", network->hidden_biases[i]);
    }

  printf("\n\nOutput weights\n");
      
  for (i = 0; i < network->nb_output_units; i++)
    {
      printf("[{");
      for (j = 0; j < network->nb_hidden_units; j++)
	printf("%f ", network->output_weights[i][j]);
      printf("}, {%f}]", network->output_biases[i]);
    }
  printf("\n");
  printf("*******************************\n");
}
