/******************************************************************************
** Project: neural_network
** Filename: rprop.c
*******************************************************************************
** Authors: A. BAILLARD & T. BAUDLOT
** Creation date: 14/07/03
** Last modification date: 25/11/03
*******************************************************************************
** Comments:
** 14/07/03	AB	Functions for rprop algorithm
**				- rprop()
** 15/07/03	TB	Rprop implementation
** 16/07/03	TB	Creation of an output file that holds energy for 
**			each iteration
** 14/04/04	AB	Add a function to handle gsl matrices as input
**			Modify stop conditions for gsl_rprop *only*
** 25/11/04	AB	Add signal in stop condition for sl_rprop *only*
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_matrix.h>
#include "network.h"
#include "prop.h"

extern short	sig_catched;

static	char	compute_chars[4] = {'/', '-', '\\', '|'};

/**
 ** rprop_rule() get the weight variation according to the rprop rule
 ** parameters:
 **	network: network being trained
 **     grad_e_i_j: current gradients
 **     prev_grad_e_i_j: previous iteration gradients
 **     double *delta_i_j: weights variations
 **     double *delta_temp_i_j: deltas (rprop specific)
 ** return value:
 **	none
 **/
static double	rprop_rule(t_network *network, 
			   double *grad_e_i_j, 
			   double *prev_grad_e_i_j,
			   double *delta_i_j,
			   double *delta_temp_i_j)
{
  double prev_delta_i_j;
  double prev_delta_temp_i_j;

  prev_delta_i_j = *delta_i_j;
  prev_delta_temp_i_j = *delta_temp_i_j;

  if ((*grad_e_i_j) * (*prev_grad_e_i_j) > 0)
    {
      *delta_temp_i_j = minimum(prev_delta_temp_i_j * N_PLUS, DELTA_MAX);
      *delta_i_j = -sign(*grad_e_i_j) * (*delta_temp_i_j);
      return *delta_i_j;
    }
  else if ((*grad_e_i_j) * (*prev_grad_e_i_j) < 0)
    {
      *delta_temp_i_j = maximum(prev_delta_temp_i_j *N_MINUS, DELTA_MIN);
      *grad_e_i_j = 0;
      return -prev_delta_i_j;
    }
  else 
    {
      *delta_i_j = - sign(*grad_e_i_j) * (*delta_temp_i_j);
      return *delta_i_j;
    }
}

/**
 ** update_network() computes new weights and biases
 ** parameters:
 **	network: network being trained
 **     vectors: set of vectors used by rprop algorithm
 ** return value:
 **	none
 **/
static void	update_network(t_network *network, t_rprop_vec *vectors)
{
  unsigned	i;
  unsigned	j;
  
  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	{
	  network->hidden_weights[i][j] += 
	    rprop_rule(network,
		       &(vectors->e_g_h_w[CURRENT][i][j]), 
		       &(vectors->e_g_h_w[PREVIOUS][i][j]),
		       &(vectors->d_h_w[i][j]),
		       &(vectors->d_t_h_w[i][j]));
	}
      network->hidden_biases[i] +=
	rprop_rule(network,
		   &(vectors->e_g_h_b[CURRENT][i]), 
		   &(vectors->e_g_h_b[PREVIOUS][i]),
		   &(vectors->d_h_b[i]),
		   &(vectors->d_t_h_b[i]));
    }
  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	{
	  network->output_weights[i][j] += 
	    rprop_rule(network,
		       &(vectors->e_g_o_w[CURRENT][i][j]), 
		       &(vectors->e_g_o_w[PREVIOUS][i][j]),
		       &(vectors->d_o_w[i][j]),
		       &(vectors->d_t_o_w[i][j]));
	}
      network->output_biases[i] += 
	rprop_rule(network,
		   &(vectors->e_g_o_b[CURRENT][i]), 
		   &(vectors->e_g_o_b[PREVIOUS][i]),
		   &(vectors->d_o_b[i]),
		   &(vectors->d_t_o_b[i]));
    }
}


/**
 ** rprop() trains the network on a set of examples 
 ** composed of an input and an expected output
 ** parameters:
 **	network: allocated network to train
 **	base_dim: dimension of the examples base
 **	input_base: set of input vectors
 **	n_input: size of an input vector
 **	expected_base: set of output vectors corresponding to input vectors
 **	n_output: size of an expected output vector
 ** return value:
 **	error code
 **/
int		rprop(t_network *network, unsigned base_dim,
		      double **input_base, unsigned n_input,
		      double **expected_base, unsigned n_output)
{
  unsigned	sample_cnt;
  
  t_rprop_vec	vectors;	
  
  double	energy;  
  double	previous_energy;
  unsigned	iter;
  unsigned	nb_iter;
  FILE		*file;

  create_rprop_vectors(network, &vectors);

  init_rprop_vectors(network, &vectors);
  iter = 0;
  nb_iter = 0;
  energy = 0;
  file = fopen("rprop.res","w");

  do
    {
      previous_energy = energy;
      energy = 0;
      init_current_grads(network, &vectors);
      
      for (sample_cnt = 0; sample_cnt < base_dim; sample_cnt++)
	{
	  feed_forward(network, input_base[sample_cnt]);
	  
	  compute_gradients(network, expected_base[sample_cnt]);
	  
	  compute_energy_grads(network, &vectors, input_base[sample_cnt]);
	  
	  compute_energy(network, expected_base[sample_cnt]);
	  energy += network->energy;
	}
      energy /= base_dim;

      update_network(network, &vectors);

      save_energy_grads(network, &vectors);

      fprintf(file, "%d\t%f\n", iter, energy);
      fflush(file);
      iter++;
      if ((float)energy != (float)previous_energy)
	nb_iter = 0;
      else
	nb_iter++;
    }
  while (energy > network->energy_treshold && nb_iter < network->max_iter);

  fclose(file);

  delete_rprop_vectors(network, &vectors);

  return LIBNN_NO_ERROR;
}

/**
 ** gsl_rprop() trains the network on a set of examples 
 ** composed of an input and an expected output
 ** parameters:
 **	network: allocated network to train
 **	input: matrix containing input data
 **	input: matrix containing expected output data
 ** return value:
 **	error code
 **/
int		gsl_rprop(t_network *network, unsigned nb_samples,
			  const gsl_matrix_float *input, const gsl_matrix_float *output)
{
  unsigned	sample_cnt;
  t_rprop_vec	vectors;	
  double	energy;  
  double	previous_energy;
  unsigned	iter;
  unsigned	nb_iter;
  unsigned	cnt;
  FILE		*file;
  double	*dbl_input;
  float		*flt_input;
  double	*dbl_output;
  float		*flt_output;
  char		round[11];

  create_rprop_vectors(network, &vectors);
  init_rprop_vectors(network, &vectors);
  iter = 0;
  nb_iter = 0;
  energy = 0;
  file = fopen("rprop.res","w");
  dbl_input = malloc(input->size2 * sizeof(double));
  dbl_output = malloc(output->size2 * sizeof(double));

  fprintf(file, "Max iter\tTreshold\n");
  fprintf(file, "%d\t%10.8f\n", network->max_iter, network->energy_treshold);
  fprintf(file, "\nIter\tMax iter\tEnergy\tDEnergy\n");
  do
    {
      previous_energy = energy;
      energy = 0;
      init_current_grads(network, &vectors);
      
      for (sample_cnt = 0; sample_cnt < nb_samples; sample_cnt++)
	{
	  flt_input = gsl_matrix_float_ptr(input, sample_cnt, 0);
	  flt_output = gsl_matrix_float_ptr(output, sample_cnt, 0);
	  for (cnt = 0; cnt < input->size2; ++cnt)
	    dbl_input[cnt] = flt_input[cnt];
	  for (cnt = 0; cnt < output->size2; ++cnt)
	    {
	      dbl_output[cnt] = flt_output[cnt];
/* 	      if (dbl_output[cnt] != 0) */
/* 		printf("%f %f\n", dbl_input[cnt], dbl_output[cnt]); */
/* 	      printf("\n"); */
	    }

	  feed_forward(network, dbl_input);
	  compute_gradients(network, dbl_output);
	  compute_energy_grads(network, &vectors, dbl_input);
	  compute_energy(network, dbl_output);
	  energy += network->energy;
	}
      energy /= input->size1;

      update_network(network, &vectors);
      save_energy_grads(network, &vectors);

      sprintf(round, "%10.8f", previous_energy);
      previous_energy = atof(round);
      sprintf(round, "%10.8f", energy);
      energy = atof(round);

      fprintf(file, "%d\t%d\t%10.8f\t%10.8f\n", iter, nb_iter, energy, 
	      xfabs(energy - previous_energy));
      fflush(file);
      printf("Computing %c (C-C to stop)\n\033[1A\033[1M", compute_chars[iter % 4]);
      ++iter;
      if (xfabs(energy - previous_energy) < 1e-8)
	++nb_iter;
      else
	nb_iter = 0;
    }
  while ((energy > network->energy_treshold) && (nb_iter < network->max_iter) && !sig_catched);

  free(dbl_input);
  free(dbl_output);
  fclose(file);
  delete_rprop_vectors(network, &vectors);
  return LIBNN_NO_ERROR;
}
