/******************************************************************************
** Project: neural_network
** Filename: rprop_vec.c
*******************************************************************************
** Authors: A. BAILLARD & T. BAUDLOT
** Creation date: 18/07/03
** Last modification date: 18/07/03
*******************************************************************************
** Comments:
** 18/07/03	AB	Functions for rprop vectors management
**				- create_rprop_vectors()
**				- delete_rprop_vectors()
**				- init_rprop_vectors()
**				- init_current_grads()
**				- compute_energy_grads()
**				- save_energy_grads()
******************************************************************************/

#include <stdlib.h>
#include "network.h"

/**
 ** create_four_vectors() creates vectors
 ** parameters:
 **	network: network being trained
 **	dbl_hidden_vector: a double vector of hidden layer
 **     dbl_output_vector: a double vector of output layer
 **     spl_hidden_vector: a simple vector of hidden layer
 **     spl_output_vector: a simple vector of output layer
 **	vectors: rprop specific vectors
 ** return value:
 **	none
 **/
static void	create_four_vectors(t_network *network,
				    double	***dbl_hidden_vector,
				    double	***dbl_output_vector,
				    double	**spl_hidden_vector,
				    double	**spl_output_vector)
{
  unsigned i;
  
  *spl_output_vector = (double *)xmalloc(network->nb_output_units * 
						 sizeof(double));

  *spl_hidden_vector = (double *)xmalloc(network->nb_hidden_units * 
						 sizeof(double));


  *dbl_output_vector = (double **)xmalloc(network->nb_output_units * 
					    sizeof(double *));
  for (i = 0; i < network->nb_output_units; i++)
    (*dbl_output_vector)[i] = (double *)xmalloc(network->nb_hidden_units * 
						 sizeof(double));
  
  *dbl_hidden_vector = (double **)xmalloc(network->nb_hidden_units * 
					     sizeof(double *));
  for (i = 0; i < network->nb_hidden_units; i++)
    (*dbl_hidden_vector)[i] = (double *)xmalloc(network->nb_input_units * 
						 sizeof(double));

}
			    

/**
 ** create_rprop_vectors() creates vectors for rprop algorithm
 ** parameters:
 **	network: network being trained
 **	vectors: rprop specific vectors
 ** return value:
 **	none
 **/
void	create_rprop_vectors(t_network *network, t_rprop_vec *vectors)
{
  create_four_vectors(network, 
		      &(vectors->e_g_h_w[CURRENT]), &(vectors->e_g_o_w[CURRENT]),
		      &(vectors->e_g_h_b[CURRENT]), &(vectors->e_g_o_b[CURRENT]));
  create_four_vectors(network, 
		      &(vectors->e_g_h_w[PREVIOUS]), &(vectors->e_g_o_w[PREVIOUS]),
		      &(vectors->e_g_h_b[PREVIOUS]), &(vectors->e_g_o_b[PREVIOUS]));

  create_four_vectors(network,
		      &(vectors->d_t_h_w), &(vectors->d_t_o_w),
		      &(vectors->d_t_h_b), &(vectors->d_t_o_b));

  create_four_vectors(network,
		      &(vectors->d_h_w), &(vectors->d_o_w),
		      &(vectors->d_h_b), &(vectors->d_o_b));
}

/**
 ** free_four_vectors() frees vectors
 ** parameters:
 **	network: network being trained
 **	dbl_hidden_vector: a double vector of hidden layer
 **     dbl_output_vector: a double vector of output layer
 **     spl_hidden_vector: a simple vector of hidden layer
 **     spl_output_vector: a simple vector of output layer
 **	vectors: rprop specific vectors
 ** return value:
 **	none
 **/
static void	free_four_vectors(t_network *network,
				    double	**dbl_hidden_vector,
				    double	**dbl_output_vector,
				    double	*spl_hidden_vector,
				    double	*spl_output_vector)
{
  unsigned i;
  
  free(spl_output_vector);
  free(spl_hidden_vector);

  for (i = 0; i < network->nb_output_units; i++)
    free(dbl_output_vector[i]);
  free(dbl_output_vector);
  
  for (i = 0; i < network->nb_hidden_units; i++)
    free(dbl_hidden_vector[i]);
  free(dbl_hidden_vector);
}
	

/**
 ** delete_rprop_vectors() frees vectors for rprop algorithm
 ** parameters:
 **	network: network being trained
 **	vectors: rprop specific vectors
 ** return value:
 **	none
 **/
void	delete_rprop_vectors(t_network *network, t_rprop_vec *vectors)
{
  free_four_vectors(network, 
		    vectors->e_g_h_w[CURRENT], vectors->e_g_o_w[CURRENT],
		    vectors->e_g_h_b[CURRENT], vectors->e_g_o_b[CURRENT]);
  free_four_vectors(network, 
		    vectors->e_g_h_w[PREVIOUS], vectors->e_g_o_w[PREVIOUS],
		    vectors->e_g_h_b[PREVIOUS], vectors->e_g_o_b[PREVIOUS]);

  free_four_vectors(network,
		    vectors->d_t_h_w, vectors->d_t_o_w,
		    vectors->d_t_h_b, vectors->d_t_o_b);

  free_four_vectors(network,
		    vectors->d_h_w, vectors->d_o_w,
		    vectors->d_h_b, vectors->d_o_b);
}


/**
 ** init_energy_grads() inits weights and biases to zero
 ** parameters:
 **	network: network being trained
 **     energy_grad_hidden_weights: vector of hidden weights variations
 **     energy_grad_output_weights: vector of output weights variations
 **     energy_grad_hidden_biases: vector of hidden biases variations
 **     energy_grad_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void    	init_energy_grads(t_network *network, 
				  double	**energy_grad_hidden_weights,
				  double	**energy_grad_output_weights,
				  double	*energy_grad_hidden_biases,
				  double	*energy_grad_output_biases)
{
  unsigned	i;
  unsigned	j;

  for (i = 0; i < network->nb_output_units; i++)
    {
      energy_grad_output_biases[i] = 0;
      for (j = 0; j < network->nb_hidden_units; j++)
	energy_grad_output_weights[i][j] = 0;
    }

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      energy_grad_hidden_biases[i] = 0;
      for (j = 0; j < network->nb_input_units; j++)
	energy_grad_hidden_weights[i][j] = 0;
    }
}

/**
 ** init_delta_temp() inits weights and biases to DELTA_0
 ** parameters:
 **	network: network being trained
 **     delta_temp_hidden_weights: vector of hidden weights variations
 **     delta_temp_output_weights: vector of output weights variations
 **     delta_temp_hidden_biases: vector of hidden biases variations
 **     delta_temp_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
static void	init_delta_temp(t_network *network,
				double	**delta_temp_hidden_weights,
				double	**delta_temp_output_weights,
				double	*delta_temp_hidden_biases,
				double	*delta_temp_output_biases)
{
  unsigned	i;
  unsigned	j;
  
  for (i = 0; i < network->nb_output_units; i++)
    {
      delta_temp_output_biases[i] = DELTA_0;
      for (j = 0; j < network->nb_hidden_units; j++)
	delta_temp_output_weights[i][j] = DELTA_0;
    }

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      delta_temp_hidden_biases[i] = DELTA_0;
      for (j = 0; j < network->nb_input_units; j++)
	delta_temp_hidden_weights[i][j] = DELTA_0;
    }
}

void	init_rprop_vectors(t_network *network, t_rprop_vec *vectors)
{
  init_energy_grads(network, vectors->e_g_h_w[PREVIOUS], vectors->e_g_o_w[PREVIOUS], 
			     vectors->e_g_h_b[PREVIOUS], vectors->e_g_o_b[PREVIOUS]);

  init_delta_temp(network, vectors->d_t_h_w, vectors->d_t_o_w,
			   vectors->d_t_h_b, vectors->d_t_o_b);
}


/**
 ** init_current_grads() init current gradients
 ** parameters:
 **	network: network being trained
 **	vectors: set of rprop vectors
 ** return value:
 **	none
 **/
void		init_current_grads(t_network *network, t_rprop_vec *vectors)
{
  init_energy_grads(network,
		    vectors->e_g_h_w[CURRENT], vectors->e_g_o_w[CURRENT],
		    vectors->e_g_h_b[CURRENT], vectors->e_g_o_b[CURRENT]);
}


/**
 ** compute_energy_grads() computes variations of weights and biases
 ** parameters:
 **	network: network being trained
 **     energy_grad_hidden_weights: vector of hidden weights variations
 **     energy_grad_output_weights: vector of output weights variations
 **     energy_grad_hidden_biases: vector of hidden biases variations
 **     energy_grad_output_biases: vector of output biases variations
 **	input_sample: sample currently used for training
 ** return value:
 **	none
 **/
void		compute_energy_grads(t_network *network, 
				     t_rprop_vec *vectors,
				     double	*input_sample)
{
  unsigned	i;
  unsigned	j;

  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	{
	  vectors->e_g_h_w[CURRENT][i][j] += 
	    network->hidden_energy[i] * input_sample[j];
	}
      vectors->e_g_h_b[CURRENT][i] += network->hidden_energy[i];
    }
  
  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	{
	  vectors->e_g_o_w[CURRENT][i][j] +=
	    network->output_energy[i] * network->hidden_values[j];
	}
      vectors->e_g_o_b[CURRENT][i] += network->output_energy[i];      
    }
}


/**
 ** save_energy_grads() save gradients at the previous time
 ** parameters:
 **	network: network being trained
 **     energy_grad_hidden_weights: vector of hidden weights variations
 **     energy_grad_output_weights: vector of output weights variations
 **     energy_grad_hidden_biases: vector of hidden biases variations
 **     energy_grad_output_biases: vector of output biases variations
 ** return value:
 **	none
 **/
void		save_energy_grads(t_network *network, t_rprop_vec *vectors)
{
  unsigned	i;
  unsigned	j;
  
  for (i = 0; i < network->nb_hidden_units; i++)
    {
      for (j = 0; j < network->nb_input_units; j++)
	vectors->e_g_h_w[PREVIOUS][i][j] = vectors->e_g_h_w[CURRENT][i][j];
      vectors->e_g_h_b[PREVIOUS][i] = vectors->e_g_h_b[CURRENT][i];
    }

  for (i = 0; i < network->nb_output_units; i++)
    {
      for (j = 0; j < network->nb_hidden_units; j++)
	vectors->e_g_o_w[PREVIOUS][i][j] = vectors->e_g_o_w[CURRENT][i][j];
      vectors->e_g_o_b[PREVIOUS][i] = vectors->e_g_o_b[CURRENT][i];
    }
}
