///////////////////////////////////////////////////////////////////////////////
// Project: defectpix 
// Filename: main.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 14/01/04
// Last modification date: 16/03/04
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 14/01/04	Creation
// 24/02/04	Add options management
// 02/03/04	Add mode management
// 11/03/04	Add computing mode
// 16/03/04	Move from defectix.cc to main.cc
///////////////////////////////////////////////////////////////////////////////

#include <string>
#include <iostream>
#include "trainer.hh"
#include "computer.hh"
#include "halosimulator.hh"
#include "trailsimulator.hh"

using namespace DefectixOptions;

static struct option const long_opts[] =
  {
    {"help",			no_argument,		0,	       	opt_help},
    {"version",			no_argument,		0,	       	opt_version},
    {"default",			no_argument,		0,	       	opt_default},
    {"mode",			required_argument,	0,		opt_mode},
    {"conf-file",		required_argument,	0,		opt_conf},
    {"input-file",		required_argument,	0,       	opt_input},
    {"mask-file",		required_argument,	0,       	opt_mask},
    {"output-file",		required_argument,	0,       	opt_output},
    {"evec-file",		required_argument,	0,       	opt_efile},
    {"nn-file",			required_argument,	0,       	opt_nnfile},
    {"rebin-factor",		required_argument,	0,       	opt_rebin},
    {"substract-bg",		required_argument,	0,       	opt_subbg},
    {"reduce-dynamic",		required_argument,	0,       	opt_dyn},
    {"bloc-size",		required_argument,	0,       	opt_bsize},
    {"nb-pc",			required_argument,	0,       	opt_nbpc},
    {"zoom-step",		required_argument,	0,       	opt_zoom},
    {"print-evec",		no_argument,		0,       	opt_evec},
    {"print-rebuilt",		no_argument,		0,       	opt_rebuilt},
    {"adder-type",		required_argument,	0,		opt_addertype},
    {"intensity",		required_argument,	0,		opt_intensity},
    {"nb-defect",		required_argument,	0,		opt_nbdef},
    {"energy-treshold",		required_argument,	0,		opt_energy},
    {"mask-treshold",		required_argument,	0,		opt_mtreshold},
    {"max-iteration",		required_argument,	0,		opt_maxiter},
    {"max-prior",		required_argument,	0,		opt_maxprior},
    {"min-prior",		required_argument,	0,		opt_minprior},
    {"non-prior",		required_argument,	0,		opt_nonprior},
    {"verbose",			no_argument,		0,		opt_verbose},
    {"nb-vector",		required_argument,	0,		opt_nbvec},
    {NULL, 0, NULL, 0}
  };

static const char short_opts[] = "hVdM:C:I:K:L:O:N:r:b:p:z:ERS:D:i:a:f:vt:e:m:x:n:l:s:";

int		main(int argc, char **argv)
{
  int		c;
  Options	options;
  Defectix	*defect = NULL;

  while ((c = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1)
    set_option(c, options);
  
  if (options.inputFile.empty())
    {
      std::cerr << ERROR_START << "no input file" << std::endl;
      return 1;
    }

  switch (options.mode)
    {
    case TRAINING:
      {
	defect = new DefectTrainer(options);
	break;
      }
    case COMPUTING:
      {
	defect = new DefectComputer(options);
	break;
      }
    case HALO:
      {
	defect = new HaloSimulator(options);
	break;
      }
    case TRAIL:
      {
	defect = new TrailSimulator(options);
	break;
      }
    default:
      {
	std::cerr << ERROR_START << "no mode chosen" << std::endl;
	return 0;
      }
    }
  
  if (defect->getError())
    return defect->printError();

  defect->exec();

  if (defect->getError())
    return defect->printError();

  delete defect;
    
  return 0;
}
