///////////////////////////////////////////////////////////////////////////////
// Project: defectpix 
// Filename: pca.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 18/03/04
// Last modification date: 13/05/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 18/03/04	Copy code from pca.hh
// 22/10/04	Add computeRebuilt()
// 21/01/05	Change writeEigenVectors() to create a MEF file
// 09/05/05	Write computeCovar() instead of calling gsl_multifit_covar()
// 13/05/05	remove computeCovar() and copy computePCA() from basix.0.3.0
///////////////////////////////////////////////////////////////////////////////

#include "pca.hh"

DefectPCA::DefectPCA(const DefectixOptions::Options& options) :
  ErrorMgr(options.verbose)
{
  _printEvec = options.printEvec;
  _blocSize = options.blocSize;
  _zoomStep = options.zoomStep;
  _eval = NULL;
  _evec = NULL;
  _pc = NULL;
  _evecFile = options.evecFile;
}

/////////////////////////////////////////////////////////////////////////////
// PCA methods								   //
/////////////////////////////////////////////////////////////////////////////
void	DefectPCA::setEigenVectors(unsigned size)
{
  if (_eval != NULL && _eval->size != size)
    {
      gsl_vector_float_free(_eval);
      gsl_matrix_float_free(_evec);
      _eval = NULL;
      _evec = NULL;
    }
  if (_eval == NULL)
    {
      _eval = gsl_vector_float_alloc(size);
      _evec = gsl_matrix_float_alloc(size, size);
    }
  if (_eval == NULL || _evec == NULL)
    setError("DefectPCA::setEigenVectors", MALLOC);
}

/////////////////////////////////////////////////////////////////////////////
errorCodes	DefectPCA::eigenVectors(const gsl_matrix_float* const const_m)
{
  gsl_matrix	*m;
  gsl_vector	*v;
  
  // compute smallest covariance matrix
  if (const_m->size1 > const_m->size2)
    {
      if ((m = gsl_matrix_alloc(const_m->size1, const_m->size2)) == NULL)
	  return setError("DefectPCA::eigenVectors", MALLOC);
      for (unsigned int cnt1 = 0; cnt1 < m->size1; ++cnt1)
	for (unsigned int cnt2 = 0; cnt2 < m->size2; ++cnt2)
	  gsl_matrix_set(m, cnt1, cnt2, gsl_matrix_float_get(const_m, cnt1, cnt2));

      // norm each image
      v = gsl_vector_alloc(m->size2);
      if (v == NULL)
	return setError("DefectPCA::eigenVectors", MALLOC);
      double variance;
      for (unsigned int cnt = 0; cnt < m->size1; cnt++)
	{
	  gsl_matrix_get_row(v, m, cnt);
	  variance = sqrt(gsl_stats_variance(v->data, 1, v->size));
	  if (fabs(variance) > 10e-5)
	    for (unsigned int cnt2 = 0; cnt2 < m->size2; cnt2++)
	      gsl_matrix_set(m, cnt, cnt2, gsl_matrix_get(m, cnt, cnt2) / variance);
	}
      gsl_vector_free(v);
    }
  else
    {
      printMessage("changing from signals space to attributes space");
      if ((m = gsl_matrix_alloc(const_m->size2, const_m->size1)) == NULL)
	return setError("DefectPCA::eigenVectors", MALLOC);
      for (unsigned int cnt1 = 0; cnt1 < m->size1; ++cnt1)
	for (unsigned int cnt2 = 0; cnt2 < m->size2; ++cnt2)
	  gsl_matrix_set(m, cnt1, cnt2, gsl_matrix_float_get(const_m, cnt2, cnt1));

      // norm each image
      v = gsl_vector_alloc(m->size1);
      if (v == NULL)
	return setError("DefectPCA::eigenVectors", MALLOC);
      double variance;
      for (unsigned int cnt = 0; cnt < m->size2; cnt++)
	{
	  gsl_matrix_get_col(v, m, cnt);
	  variance = sqrt(gsl_stats_variance(v->data, 1, v->size));
	  if (fabs(variance) > 10e-5)
	    for (unsigned int cnt2 = 0; cnt2 < m->size1; cnt2++)
	      gsl_matrix_set(m, cnt2, cnt, gsl_matrix_get(m, cnt2, cnt) / variance);
	}
      gsl_vector_free(v);
    }

  // set environment to compute covariance matrix
  setEigenVectors(m->size2);
  if (getError() != NO_ERROR)
    return getError();

  gsl_vector	*eval = gsl_vector_alloc(m->size2);
  gsl_matrix	*evec = gsl_matrix_alloc(m->size2, m->size2);
  gsl_matrix	*covar = gsl_matrix_alloc(m->size2, m->size2);
  if (eval == NULL || evec == NULL || covar == NULL)
    return setError("DefectPCA::eigenVectors", MALLOC);
  printMessage("computational environment OK");

  // compute covariance matrix
  gsl_multifit_covar(m, 0.0, covar);
  printMessage("covariance matrix computation OK");

  // compute eigen vectors and eigen values of covariance matrix
  gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc(const_m->size2);
  if (w == NULL)
    return setError("DefectPCA::eigenVectors", MALLOC);
  gsl_eigen_symmv(covar, eval, evec, w);
  gsl_eigen_symmv_free(w);
  gsl_matrix_free(covar);
  gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_VAL_ASC);

  // if m was transposed t attributes space, turn back to original space
  if (const_m->size1 < const_m->size2)
    {
      gsl_matrix	*evec2;

      printMessage("turning back to signals space");
      gsl_matrix_float_free(_evec);
      evec2 = gsl_matrix_alloc(m->size1, m->size2);
      _evec = gsl_matrix_float_alloc(m->size1, m->size2);

      // Compute _evec = m evec, that is the eigen vectors in the original space
      gsl_blas_dgemm(CblasNoTrans, CblasNoTrans,
		     1.0, m, evec,
		     0.0, evec2);
      gsl_matrix_free(evec);
      evec = evec2;

      // Norm new eigen vectors by eigen values
      double val;
      for (unsigned colCnt = 0; colCnt < evec->size2; colCnt++)
	{
	  gsl_vector_view v = gsl_matrix_column(evec, colCnt);
	  val = sqrt(gsl_vector_get(eval, colCnt));
	  gsl_vector_scale(&v.vector, val);
	}
    }

  // conversion from double to float
  for (unsigned int cnt1 = 0; cnt1 < eval->size; ++cnt1)
    {
      gsl_vector_float_set(_eval, cnt1, gsl_vector_get(eval, cnt1));
      for (unsigned int cnt2 = 0; cnt2 < evec->size1; ++cnt2)
	gsl_matrix_float_set(_evec, cnt2, cnt1, gsl_matrix_get(evec, cnt2, cnt1));
    }

  gsl_matrix_free(m);
  gsl_matrix_free(evec);
  gsl_vector_free(eval);

  printMessage("eigen vectors and eigen values computed");
  return setError("DefectPCA::eigenVectors", NO_ERROR);
}  

/////////////////////////////////////////////////////////////////////////////
// public function to call						   //
/////////////////////////////////////////////////////////////////////////////
void	DefectPCA::computePCA(gsl_matrix_float *m, const int nb_pc)
{
  if (m == NULL)
    setError("DefectPCA::computePCA", NOT_ALLOC_MAT);
  else if (_evec == NULL)
    setError("DefectPCA::computePCA", eigenVectors(m));

  if (getError() == NO_ERROR)
    {
      if ((_pc == NULL) && (_pc = gsl_matrix_float_alloc(m->size1, nb_pc)) == NULL)
	setError("DefectPCA::computePCA", MALLOC);
      else
	{
	  unsigned cnt;
	  gsl_matrix_float_view evecv = gsl_matrix_float_submatrix(_evec, 0, 0, _evec->size1, _pc->size2);
	  for (cnt = 0; (cnt + 10) < m->size1; cnt += 10)
	    {
	      gsl_matrix_float_view mv = gsl_matrix_float_submatrix(m, cnt, 0, 10, m->size2);
	      gsl_matrix_float_view pcv = gsl_matrix_float_submatrix(_pc, cnt, 0, 10, _pc->size2);
	      // Compute _pc = m evec, that is the principal components of the image
	      gsl_blas_sgemm(CblasNoTrans, CblasNoTrans,
			     1.0, &mv.matrix, &evecv.matrix,
			     0.0, &pcv.matrix);
	    }
	  gsl_matrix_float_view mv = gsl_matrix_float_submatrix(m, cnt, 0, m->size1 - cnt, m->size2);
	  gsl_matrix_float_view pcv = gsl_matrix_float_submatrix(_pc, cnt, 0, _pc->size1 - cnt, _pc->size2);
	  // Compute _pc = m evec, that is the principal components of the image
	  gsl_blas_sgemm(CblasNoTrans, CblasNoTrans,
			 1.0, &mv.matrix, &evecv.matrix,
			 0.0, &pcv.matrix);
	  if (_printEvec == true)
	    writeEigenVectors(_evec);
	}
    }
}

void	DefectPCA::computePCA(gsl_matrix_float *m, const int nb_pc, gsl_matrix_float *evec)
{
  _evec = evec;
  computePCA(m, nb_pc);
}

/////////////////////////////////////////////////////////////////////////////
// output methods							   //
/////////////////////////////////////////////////////////////////////////////
void		DefectPCA::writeEigenVectors(gsl_matrix_float *evec)
{
  FILE			*f;
  fitsfile		*ptr;
  long			fpixel[2];
  long			blocSizes[2];
  int			status(0);
  gsl_vector_float	*v1;
  unsigned short	value;
  int			key;

  blocSizes[0] = blocSizes[1] = _blocSize;
  fpixel[0] = fpixel[1] = 1;
  if ((v1 = gsl_vector_float_alloc(evec->size1)) == NULL)
    setError("DefectPCA::writeEigenVectors", MALLOC);
  else
    {
      fits_create_file(&ptr, "!evec.fits", &status);
      fits_create_hdu(ptr, &status);
      value = TRUE;
      fits_write_key(ptr, TLOGICAL, "SIMPLE", &value, NULL, &status);
      key = 16;
      fits_write_key(ptr, TINT, "BITPIX", &key, NULL, &status);
      key = 0;
      fits_write_key(ptr, TINT, "NAXIS", &key, NULL, &status);
      fits_write_key(ptr, TLOGICAL, "EXTEND", &value, NULL, &status);
      fits_write_key(ptr, TINT, "NEXTEND", &_evec->size2, NULL, &status);
      for (unsigned int cnt = 0; cnt < evec->size2; cnt++)
	{
	  fits_create_img(ptr, FLOAT_IMG, 2, blocSizes, &status);
	  gsl_matrix_float_get_col(v1, evec, cnt);
	  fits_write_pix(ptr, TFLOAT, fpixel, blocSizes[0] * blocSizes[1], 
			 gsl_vector_float_ptr(v1, 0), 
			 &status);
	  setFitsError(status);
	}
      fits_close_file(ptr, &status);
      gsl_vector_float_free(v1);
      printMessage("principal components written in evec.fits (fits format)");

      if ((f = fopen(_evecFile.c_str(), "w")) != NULL)
	{
	  char	msg[_evecFile.size() + 50];
	  fwrite(&evec->size1, sizeof(unsigned long), 1, f);
	  fwrite(&evec->size2, sizeof(unsigned long), 1, f);
	  gsl_matrix_float_fwrite(f, evec);
	  fclose(f);
	  sprintf(msg, "principal components written in %s (gsl format)", _evecFile.c_str());
	  printMessage(msg);
	}
      else
	setError("pca no good code", FITS);
    }
}

/////////////////////////////////////////////////////////////////////////////
void		DefectPCA::computeRebuilt(const gsl_matrix_float *src, gsl_matrix_float *dest, int nb_pc)
{
  gsl_vector_float	*v;
  gsl_matrix_float	*pc;
  gsl_matrix_float	*evect;
  gsl_matrix_float_view evecv;

  v = gsl_vector_float_alloc(_blocSize * _blocSize);
  pc = gsl_matrix_float_alloc(1, nb_pc);
  evecv = gsl_matrix_float_submatrix(_evec, 0, 0, _evec->size1, nb_pc);
  evect = gsl_matrix_float_alloc(nb_pc, _evec->size1);
  gsl_matrix_float_transpose_memcpy(evect, &evecv.matrix);

  for (unsigned row = 0; row < src->size1; ++row)
    for (unsigned col = 0; col < src->size2; ++col)
      {
	getVectorfromBloc(v, src, row, col);
	gsl_matrix_float_view mv = gsl_matrix_float_view_vector(v, 1, v->size);
	// Compute _pc = mv evec, that is the principal components of the image
	gsl_blas_sgemm(CblasNoTrans, CblasNoTrans,
		       1.0, &mv.matrix, &evecv.matrix,
		       0.0, pc);
	
	// Compute mv = pc evec'
	gsl_blas_sgemm(CblasNoTrans, CblasNoTrans,
		       1.0, pc, evect,
		       0.0, &mv.matrix);
	setBlocfromVector(dest, row, col, v);
      }
  gsl_matrix_float_free(pc);
  gsl_vector_float_free(v);
}
