///////////////////////////////////////////////////////////////////////////////
// Project: defectpix 
// Filename: pca.hh
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 17/02/04
// Last modification date: 09/05/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 17/02/04	Creation
// 23/02/04	Modify inheritance
//		Modify PCA and print methods
//		Apply variance division in eigenVectors()
// 16/03/04	Clean code
//		Complete error management *
// 18/03/04	Move code to pca.cc
// 09/05/05	Add computeCovar()
///////////////////////////////////////////////////////////////////////////////

#ifndef PCA_HH_
# define PCA_HH_

# include <gsl/gsl_matrix_float.h>
# include <gsl/gsl_statistics_double.h>
# include <gsl/gsl_multifit_nlin.h>
# include <gsl/gsl_eigen.h>
# include <gsl/gsl_blas.h>
# include <fitsio.h>
# include "common.hh"

using namespace DefectixError;

class DefectPCA : public ErrorMgr
{
public:
  DefectPCA(const DefectixOptions::Options& options);

private:
  /////////////////////////////////////////////////////////////////////////////
  // function in pca.cc							     //
  /////////////////////////////////////////////////////////////////////////////
  void	setEigenVectors(unsigned size);

  /////////////////////////////////////////////////////////////////////////////
  errorCodes	eigenVectors(const gsl_matrix_float* const const_m);

public:
  /////////////////////////////////////////////////////////////////////////////
  // function in pca.cc							     //
  /////////////////////////////////////////////////////////////////////////////
  void	computePCA(gsl_matrix_float *m, const int nb_pc);

  /////////////////////////////////////////////////////////////////////////////
  void	computePCA(gsl_matrix_float *m, const int nb_pc, gsl_matrix_float *evec);

  /////////////////////////////////////////////////////////////////////////////
  void	computeRebuilt(const gsl_matrix_float *src, gsl_matrix_float *dest, int nb_pc);

  /////////////////////////////////////////////////////////////////////////////
  void	writeEigenVectors(gsl_matrix_float *evec);

private:
  /////////////////////////////////////////////////////////////////////////////
  inline void		printMessage(const char *msg) const
  {
    std::cout << "\033[1;32mPCA:\033[0;0m " << msg << std::endl;    
  }

public:
  /////////////////////////////////////////////////////////////////////////////
  // functions in convert.cc						     //
  /////////////////////////////////////////////////////////////////////////////
  void		getVectorfromBloc(gsl_vector_float *dest, const gsl_matrix_float * const src,
				  unsigned row, unsigned col);

  /////////////////////////////////////////////////////////////////////////////
  void		setBlocfromVector(gsl_matrix_float * dest, unsigned row, unsigned col, 
				  const gsl_vector_float * const src);

public:
  /////////////////////////////////////////////////////////////////////////////
  // accessors								     //
  /////////////////////////////////////////////////////////////////////////////
  inline const gsl_matrix_float *const	getEigenVectors() const
  {
    return _evec;
  }

  /////////////////////////////////////////////////////////////////////////////
  inline const gsl_vector_float *const	getEigenValues() const
  {
    return _eval;
  }

  /////////////////////////////////////////////////////////////////////////////
  inline const gsl_matrix_float *const	getPrincipalComponents() const
  {
    return _pc;
  }

protected:
  /////////////////////////////////////////////////////////////////////////////
  // attributes								     //
  /////////////////////////////////////////////////////////////////////////////
  int		_blocSize;
  int		_zoomStep;
  gsl_matrix_float	*_pc;
  gsl_matrix_float	*_evec;
  gsl_vector_float	*_eval;
  bool		_printEvec;
  std::string	_evecFile;
};


#endif /* !PCA_HH_ */
