///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: pretreat.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 16/03/04
// Last modification date: 22/03/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 16/03/04	Create from defectix.hh
//		Contain functions to pretreat images
//			reduceDyn()
//			substractBackground()
//			computeBackground()
//			rebin()
//			template rebin()
// 06/04/04	Move template rebin() to read.cc
// 17/16/04	Debug computeBackground() 
// 22/03/05	Debug rebin() using floor()
///////////////////////////////////////////////////////////////////////////////
 
#include "defectix.hh"

/////////////////////////////////////////////////////////////////////////////
// image pretreatment methods						   //
/////////////////////////////////////////////////////////////////////////////
double	Defectix::computeBackground(const gsl_vector_float * const pixels, 
			  const unsigned int factor, double stdev)
{
  unsigned int	range = pixels->size / factor;
  unsigned int	pos = 0;
  unsigned int	newRange;
  unsigned int	cnt;
  double	delta = gsl_vector_float_get(pixels, range) - gsl_vector_float_get(pixels, 0);
  double	tmp;

  if (range == 0)
    return gsl_vector_float_get(pixels, 0);

  for (cnt = range * 2; cnt + range < pixels->size; cnt += range)
    {
      tmp = gsl_vector_float_get(pixels, cnt) - gsl_vector_float_get(pixels, cnt - range);
      if (tmp <= delta)
	{
	  delta = tmp;
	  pos = cnt;
	}
    }
  tmp = gsl_vector_float_get(pixels, pixels->size - 1) - gsl_vector_float_get(pixels, cnt);
  if (tmp <= delta)
    pos = cnt;
  if (pos + range > pixels->size)
    newRange = pixels->size - pos;
  else
    newRange = range;
  
  float		median = gsl_vector_float_get(pixels, pos + newRange / 2);
  float		mean = gsl_stats_float_mean(&pixels->data[pos], 1, newRange);
  float		new_stdev = gsl_stats_float_sd_m(&pixels->data[pos], 1, newRange, mean);
  
  if (new_stdev - stdev > 0.2 * stdev)
    {
      gsl_vector_float_const_view subpixels = gsl_vector_float_const_subvector(pixels, pos, newRange);
      return computeBackground(&(subpixels.vector), factor, new_stdev);
    }
  else
    if (gsl_vector_float_get(pixels, pos) > mean - 3 * new_stdev &&
	gsl_vector_float_get(pixels, pos + newRange - 1) < mean + 3 * new_stdev)
      return (mean);
    else
      return (2.5 * median - 1.5 * mean);
}

/////////////////////////////////////////////////////////////////////////////
void	Defectix::substractBackground(gsl_matrix_float *m, s_ext_info *info)
{
  gsl_matrix_float_add_constant(m, -info->bg);
  info->min -= info->bg;
  info->max -= info->bg;
  info->bg = 0;
}

/////////////////////////////////////////////////////////////////////////////
void	Defectix::computeInfo(gsl_matrix_float *m, gsl_vector_float *pixels, s_ext_info *info)
{
  for (unsigned int cnt = 0; cnt < (m->size1 / 10); cnt++)
    gsl_vector_float_memcpy(&gsl_vector_float_subvector(pixels, cnt * m->size2, m->size2).vector, 
		      &gsl_matrix_float_row(m, cnt * 10).vector);
  gsl_sort_vector_float(pixels);

  info->min = gsl_vector_float_get(pixels, 0);
  info->max = gsl_vector_float_get(pixels, pixels->size - 1);
  info->bg = computeBackground(pixels, 5, 0);
}

/////////////////////////////////////////////////////////////////////////////
void	Defectix::reduceDyn(gsl_matrix_float *m, s_ext_info *info)
{
  if (getError() == NO_ERROR)
    {
//       double stdev = gsl_stats_sd(m->data, 1, m->size1 * m->size2);
      //**************** DANGER *************************//
      double stdev = 100;

      for (unsigned int row = 0; row < m->size1; row++)
	for (unsigned int col = 0; col < m->size2; col++)
	  {
	    double x = gsl_matrix_float_get(m, row, col);
	    gsl_matrix_float_set(m, row, col, (x > 0. ? 1. : -1.) * log(1. + fabs(x) / stdev) / log(1000));
	  }
      info->min = (info->min > 0. ? 1. : -1.) * log(1. + fabs(info->min) / stdev) / log(1000);
      info->max = (info->max > 0. ? 1. : -1.) * log(1. + fabs(info->max) / stdev) / log(1000);
      info->bg = (info->bg > 0. ? 1. : -1.) * log(1. + fabs(info->bg) / stdev) / log(1000);
    }
}

/////////////////////////////////////////////////////////////////////////////
void	Defectix::rebin(void **src, long bitpix, const long axes[2])
{
  if (bitpix == FLOAT_IMG)
    {
      float *dest = new (float[axes[0] * axes[1] / (_factor * _factor)]);
      rebin((float **)src, dest, axes);
    }    
  else if (bitpix == DOUBLE_IMG)
    {
      double *dest = new (double[axes[0] * axes[1] / (_factor * _factor)]);
      rebin((double **)src, dest, axes);
    }    
  else if (bitpix == BYTE_IMG)
    {
      char *dest = new (char[axes[0] * axes[1] / (_factor * _factor)]);
      rebin((char **)src, dest, axes);
    }    
  else if (bitpix == LONG_IMG)
    {
      long *dest = new (long[axes[0] * axes[1] / (_factor * _factor)]);
      rebin((long **)src, dest, axes);
    }    
  else if (bitpix == SHORT_IMG)
    {
      int *dest = new (int[axes[0] * axes[1] / (_factor * _factor)]);
      rebin((int **)src, dest, axes);
    }    
}

/////////////////////////////////////////////////////////////////////////////
template <typename T>
void	Defectix::rebin(T **src, T *dest, const long axes[2])
{
  long	destAxes[2];
  long	srcPos[2];
  long	destPos[2];
  long	srcIndex;
  long	destIndex;
  long	fac[2];
  long	div;
  
  if (dest != NULL)
    {
      destAxes[0] = floor(axes[0] / _factor);
      destAxes[1] = floor(axes[1] / _factor);
      for (srcPos[1] = 0; srcPos[1] < axes[1] ; srcPos[1] += _factor)
	{
	  destPos[1] = srcPos[1] / _factor;
	  for (srcPos[0] = 0; srcPos[0] < axes[0]; srcPos[0] += _factor)
	    {
	      destPos[0] = srcPos[0] / _factor;
	      div = 0;
	      if ((destPos[0] < destAxes[0]) && (destPos[1] < destAxes[1]))
		{
		  destIndex = destPos[1] * destAxes[0] + destPos[0];
		  for (fac[0] = 0; fac[0] < _factor; fac[0]++)
		    for (fac[1] = 0; fac[1] < _factor; fac[1]++)
		      if ((srcPos[0] +fac[0] < axes[0]) && (srcPos[1] + fac[1] < axes[1]))
			{
			  srcIndex = (srcPos[1] + fac[1]) * axes[0] + srcPos[0] + fac[0];
			  dest[destIndex] += (*src)[srcIndex];
			  div++;
			}
		  dest[destIndex] /= div;
		}
	    }
	}
      delete (*src);
      *src = dest;
    }
  else
    setError("Defectix::rebin", MALLOC);
}
