///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: read.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 16/03/04
// Last modification date: 22/03/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 16/03/04	Create from defectix.hh
//		Contain functions to read images
//			commonRead()
//			readImageExtension()
//			readAllImageExtensions()
//			unreadImageExtension()
//			unreadAllImageExtensions()
//			readMaskExtension()
//			readAllMaskExtensions()
//			unreadMaskExtension()
//			unreadAllMaskExtensions()
// 06/04/04	Modify commonRead to read image directly in a gsl_matrix_float
//		Translation should be done elsewhere (getMatrixfromImage())
// 10/01/05	Modify functions to manage single FITS files
//			commonRead(int, gsl_matrix_float *, FitsManager&)
//			readImageExtension()
//			unreadImageExtension()
//			readMaskExtension()
//			unreadMaskExtension()
// 22/03/05	Debug readImageExtension() and readMaskExtension()
//		using floor()
///////////////////////////////////////////////////////////////////////////////

#include "defectix.hh"

template <typename T>
errorCodes	Defectix::commonRead(gsl_matrix_float *m, T **src, FitsManager& mgr)
{
  if (_factor > 1)
    rebin((void **)src, mgr.getBitpix(), mgr.getAxes());
  if (getError() != NO_ERROR)
    return getError();

  if (m == NULL)
    return setError("Defectix::commonRead", MALLOC);

  for (unsigned int row = 0; row < m->size1; ++row)
    for (unsigned int col = 0; col < m->size2; ++col)
      {
	if (isnan((*src)[row * m->size2 + col]))
	  gsl_matrix_float_set(m, row, col, 0);
	else
	  gsl_matrix_float_set(m, row, col, (*src)[row * m->size2 + col]);
      }

  return setError("Defectix::commonRead", NO_ERROR);
}


errorCodes	Defectix::commonRead(int pos, gsl_matrix_float *m, FitsManager& mgr)
{
  void		*tmp = NULL;
  int		bitpix;

  if (mgr.getType() == FitsManager::NO_EXTENDED)
    mgr.uniqueHDU();
  else
    mgr.extHDU(pos);
  mgr.readImage(&tmp);
  bitpix = mgr.getBitpix();

  if (bitpix == FLOAT_IMG)
    commonRead(m, (float **)&tmp, mgr);
  else if (bitpix == DOUBLE_IMG)
    commonRead(m, (double **)&tmp, mgr);
  else if (bitpix == BYTE_IMG)
    commonRead(m, (char **)&tmp, mgr);
  else if (bitpix == LONG_IMG)
    commonRead(m, (long **)&tmp, mgr);
  else if (bitpix == SHORT_IMG)
    commonRead(m, (int **)&tmp, mgr);
  else
    return setError("Defectix::commonRead", WRONG_BITPIX);

  mgr.unreadImage(&tmp);
  return getError();
}

void		Defectix::commonPrint(int pos, const char * const type) const
{
  char	msg[128];
  sprintf(msg, 
	  "%s %d read\n(rebin %d, background substraction %s, dynamic reduction %s)", 
	  type, pos, _factor, 
	  (_substractBg ? "YES" : "NO"), 
	  (_reduceDyn ? "YES" : "NO"));
  printMessage(msg);
}

/////////////////////////////////////////////////////////////////////////////
// read image methods							   //
/////////////////////////////////////////////////////////////////////////////
errorCodes	Defectix::readImageExtension(int pos, bool createMask)
{
  gsl_matrix_float	*m = NULL;
  gsl_vector_float	*pixels = NULL;
  s_ext_info	extInfo;

  if (_imageMgr.getType() == FitsManager::NO_EXTENDED)
    _imageMgr.uniqueHDU();
  else
    _imageMgr.extHDU(pos);

  m = gsl_matrix_float_alloc(floor(_imageMgr.getAxis(1) / _factor), floor(_imageMgr.getAxis(0) / _factor));
  if (commonRead(pos, m, _imageMgr) != NO_ERROR)
    return getError();

  pixels = gsl_vector_float_alloc(m->size1 * m->size2 / 10);
  if (pixels == NULL)
    setError("Defectix::computeInfo", MALLOC);

  computeInfo(m, pixels, &extInfo);

  if (_substractBg == true)
    substractBackground(m, &extInfo);

  if (_reduceDyn == true)
    reduceDyn(m, &extInfo);
  if (getError() != NO_ERROR)
    return getError();

  _images.insert(std::pair<int, gsl_matrix_float *>(pos, m));
  _imagesInfo.insert(std::pair<int, s_ext_info>(pos, extInfo));

  if (createMask)
    {
      gsl_matrix_float *mask = gsl_matrix_float_alloc(m->size1, m->size2);
      if (mask != NULL)
	_masks.insert(std::pair<int, gsl_matrix_float *>(pos, mask));
      else
	return setError("Defectix::readImageExtension", MALLOC);
    }
  commonPrint(pos, "image");
  return getError();
}

/////////////////////////////////////////////////////////////////////////////  
errorCodes	Defectix::readAllImageExtensions(bool createMask)
{
  int		cnt;

  for (cnt = 1; cnt < _imageMgr.getNExtend(); cnt++)
    {
      if (readImageExtension(cnt, createMask) != NO_ERROR)
	return getError();
      overwriteMessage();
      overwriteMessage();
    }
  readImageExtension(cnt, createMask);
  return getError();
}

/////////////////////////////////////////////////////////////////////////////  
errorCodes	Defectix::reduceAllImagesDyn()
{
  char	msg[128];
  int	cnt;

  for (cnt = 1; cnt < _imageMgr.getNExtend(); cnt++)
    {
      reduceDyn(_images[cnt], &_imagesInfo[cnt]);
      if (getError() != NO_ERROR)
	return getError();
      sprintf(msg, 
	      "extension %d, post dynamic reduction YES", 
	      cnt);
      printMessage(msg);
      overwriteMessage();
    }

  reduceDyn(_images[cnt], &_imagesInfo[cnt]);
  if (getError() != NO_ERROR)
    return getError();
  sprintf(msg, 
	  "extension %d, post dynamic reduction YES", 
	  cnt);
  printMessage(msg);
  
  return getError();
}


/////////////////////////////////////////////////////////////////////////////
// unread image methods							   //
/////////////////////////////////////////////////////////////////////////////
void		Defectix::unreadImageExtension(int pos)
{
  if (_images[pos] != NULL)
    {
      if (_imageMgr.getType() == FitsManager::NO_EXTENDED)
	_imageMgr.uniqueHDU();
      else
	_imageMgr.extHDU(pos);
      gsl_matrix_float_free(_images[pos]);
      _images.erase(pos);
    }
}

/////////////////////////////////////////////////////////////////////////////  
void		Defectix::unreadAllImageExtensions()
{
  for (int cnt = 1; cnt <= _imageMgr.getNExtend(); cnt++)
      unreadImageExtension(cnt);
}

/////////////////////////////////////////////////////////////////////////////
// read mask methods							   //
/////////////////////////////////////////////////////////////////////////////
errorCodes	Defectix::readMaskExtension(int pos)
{
  long int	axis[2];
  gsl_matrix_float	*m = NULL;
  
  if (_imageMgr.getType() == FitsManager::NO_EXTENDED)
    {
      _imageMgr.uniqueHDU();
      _maskMgr.uniqueHDU();
    }
  else
    {
      _imageMgr.extHDU(pos);
      _maskMgr.extHDU(pos);
    }
  axis[0] = _maskMgr.getAxis(0);
  axis[1] = _maskMgr.getAxis(1);
  m = gsl_matrix_float_alloc(floor(axis[1] / _factor), floor(axis[0] / _factor));
  if (axis[0] != _imageMgr.getAxis(0) || axis[1] != _imageMgr.getAxis(1))
    return setError("Defectix::readMaskExtension", MASK_SIZE);
  if (commonRead(pos, m, _maskMgr) != NO_ERROR)
    return getError();
  _masks.insert(std::pair<int, gsl_matrix_float *>(pos, m));
  commonPrint(pos, "mask");
  return getError();
}

/////////////////////////////////////////////////////////////////////////////
errorCodes	Defectix::readAllMasks()
{
  int		cnt;
  
  for (cnt = 1; cnt < _maskMgr.getNExtend(); cnt++)
    {
      if (readMaskExtension(cnt) != NO_ERROR)
	return getError();
      overwriteMessage();
      overwriteMessage();
    }
  readMaskExtension(cnt);
  return getError();
}

/////////////////////////////////////////////////////////////////////////////
// unread mask methods							   //
/////////////////////////////////////////////////////////////////////////////
void		Defectix::unreadMaskExtension(int pos)
{
  if (_masks[pos] != NULL)
    {
      if (_maskMgr.getType() == FitsManager::NO_EXTENDED)
	_maskMgr.uniqueHDU();
      else
	_maskMgr.extHDU(pos);
      gsl_matrix_float_free(_masks[pos]);
      _masks.erase(pos);
    }
}

/////////////////////////////////////////////////////////////////////////////
void		Defectix::unreadAllMasks()
{
  for (int cnt = 1; cnt <= _maskMgr.getNExtend(); cnt++)
    unreadMaskExtension(cnt);
}
