///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: simulator.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 18/03/04
// Last modification date: 13/05/05
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 18/03/04	Copy code from simulator.hh
// 13/05/05	Modify writeAllImageExtensions() to add copyHDU()
///////////////////////////////////////////////////////////////////////////////

#include "simulator.hh"

using namespace DefectixError;

///////////////////////////////////////////////////////////////////////////////
// unique constructor							     //
///////////////////////////////////////////////////////////////////////////////
DefectSimulator::DefectSimulator(const DefectixOptions::Options& options) :
  Defectix(options)
{
  if (getError() == NO_ERROR && setError("DefectSimulator::DefectSimulator", _maskMgr.getError()) == NO_ERROR)
    {
      _nbDefects = options.nbDefects;
      _type = options.simulatorType;
      _outputMgr.copyProperties(_imageMgr);
      _intensity = options.intensity;
      if (_outputMgr.getError() != NO_ERROR)
	setError("DefectSimulator::DefectSimulator", _outputMgr.getError());
      _maskMgr.copyProperties(_imageMgr);
      if (_maskMgr.getError() != NO_ERROR)
	setError("DefectSimulator::DefectSimulator", _maskMgr.getError());
    }
}

const errorCodes	DefectSimulator::writeAllImageExtensions()
{
  char		msg[24];

  if (_imageMgr.getType() == FitsManager::NO_EXTENDED)
    if (_images[1] != NULL)
      {
	_imageMgr.uniqueHDU();
	_outputMgr.copyHDU(_imageMgr);
	_outputMgr.setHDUAsImage(FLOAT_IMG, 
			       _imageMgr.getNAxes(), 
			       _imageMgr.getAxes(), 
			       _factor);
	
	_outputMgr.writeImage((void **)&(_images[1]->data));
	sprintf(msg, "Defected extension %d written", 1);
	printMessage(msg);
      }
    else
      return setError("Defectix::writeAllMasks", NO_MASK);

  else
    {
      for (int cnt = 1; cnt <= _imageMgr.getNExtend(); ++cnt)
	{
	  if (_images[cnt] != NULL)
	    {
	      _imageMgr.extHDU(cnt);
	      _outputMgr.copyHDU(_imageMgr);
	      _outputMgr.setHDUAsImage(FLOAT_IMG, 
				       _imageMgr.getNAxes(), 
				       _imageMgr.getAxes(), 
				       _factor);
	      _outputMgr.writeImage((void **)&(_images[cnt]->data));
	      sprintf(msg, "Defected extension %d written", cnt);
	      printMessage(msg);
	      overwriteMessage();
	    }
	  else
	    return setError("DefectSimulator::writeAllImageExtensions", NO_EXT);
	}
      printMessage(msg);
    }

  return NO_ERROR;
}

///////////////////////////////////////////////////////////////////////////////
// main function							     //
///////////////////////////////////////////////////////////////////////////////
const errorCodes	DefectSimulator::exec()
{
  bool			tmpDyn = _reduceDyn;

  _reduceDyn = false;

  readAllImageExtensions(true);
  if (getError() != NO_ERROR)
    return getError();

  if (_type == DefectixOptions::EXTENSION)
    addDefectsPerExtension(_nbDefects);
  else
    addDefectsOnMosaic(_nbDefects);
  if (getError() != NO_ERROR)
    return getError();

  if (tmpDyn == true)
    {
      _reduceDyn = true;
      if (reduceAllImagesDyn() != NO_ERROR)
	return getError();
    }

  writeAllImageExtensions();
  if (getError() != NO_ERROR)
    return getError();
    
  writeAllMasks();
  if (getError() != NO_ERROR)
    return getError();
    
  unreadAllImageExtensions();
    
  return getError();
}

/////////////////////////////////////////////////////////////////////////////
void		DefectSimulator::addDefectsOnMosaic(int nb)
{
  int		extend;
  int		cnt;
    
  cnt = 0;
  do
    {
      extend = (int) (_imageMgr.getNExtend() * (rand()/(RAND_MAX + 1.0)) + 1);
      addDefectsOnExtension(1, extend);
      ++cnt;
    }
  while (cnt < nb && getError() == NO_ERROR);
}
  
/////////////////////////////////////////////////////////////////////////////
void		DefectSimulator::addDefectsPerExtension(int nb)
{
  int		extend;
  int		nExtend;
    
  extend = 1;
  nExtend = _imageMgr.getNExtend();
  do
    {
      addDefectsOnExtension(nb, extend);
      ++extend;
    }
  while (extend <= nExtend && getError() == NO_ERROR);
}

/////////////////////////////////////////////////////////////////////////////
void		DefectSimulator::addDefectsOnExtension(int nb, int extend)
{
  gsl_matrix_float	*matrices[2];

  if (_imageMgr.getType() == FitsManager::EXTENDED)
    _imageMgr.extHDU(extend);
  else
    _imageMgr.uniqueHDU();
  if (_imageMgr.getError() != NO_ERROR)
    setError("HaloSimulator::addDefectsOnExtension", _imageMgr.getError());
  else if ((matrices[0] = _images[extend]) == NULL)
    setError("HaloSimulator::addDefectsOnExtension", NO_EXTENSION);
  else if ((matrices[1] = _masks[extend]) == NULL)
    setError("HaloSimulator::addDefectsOnExtension", NO_MASK);
  else
    {
      for (int cnt = 0; cnt < nb; ++cnt)
	createDefect(matrices[0], matrices[1], extend);
	
      char	msg[24];
      sprintf(msg, "extension %d defected", extend);
      printMessage(msg);
      overwriteMessage();
    }
}
