///////////////////////////////////////////////////////////////////////////////
// Project: defectix 
// Filename: trainer.cc
// Directory: src
///////////////////////////////////////////////////////////////////////////////
// Authors: A. BAILLARD
// Creation date: 18/03/04
// Last modification date: 22/10/04
///////////////////////////////////////////////////////////////////////////////
// Comments:
// 18/03/04	Copy code from trainer.hh
// 07/04/04	Modify functions to use gsl_matrix_float and to convert images
// 17/06/04	Modify createVectors() to select random blocs
//		Modify createKLBase() to select significant blocs
// 22/10/04	Modify exec() to call writeAllRebuilts()
// 22/11/04	Correct a bug in createKLBase()
///////////////////////////////////////////////////////////////////////////////

#include "trainer.hh"

using namespace DefectixError;

///////////////////////////////////////////////////////////////////////////////
DefectTrainer::DefectTrainer(const DefectixOptions::Options& options) : 
  Defectix(options)
{
  if (getError() == NO_ERROR && setError("DefectTrainer::DefectTrainer", _maskMgr.getError()) == NO_ERROR)
    {
      _evecFile = options.evecFile;
      _nnFile = options.nnFile;
      _nbPC = options.nbPC;
      _nbVectors = options.nbVectors;
      _energyTreshold = options.energyTreshold;
      _outputMgr.copyProperties(_imageMgr);
      _maxIter = options.maxIter;
      _PCA = new DefectPCA(options);
      _printRebuilt = options.printRebuilt;
      _maxPrior = options.maxPrior;
      _minPrior = options.minPrior;
      _nonPrior = options.nonPrior;
      _l0 = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
const errorCodes	DefectTrainer::exec()
{
  readAllImageExtensions();
  if (getError() != NO_ERROR)
    return getError();

  readAllMasks();
  if (getError())
    return getError();
    
  if (train() != NO_ERROR)
    return getError();
    
  if (_printRebuilt == true)
    writeAllRebuilts(_nbPC);

  return getError();
}

///////////////////////////////////////////////////////////////////////////////
errorCodes	DefectTrainer::createVectors(gsl_matrix_float *input,
					     gsl_matrix_float *output,
					     std::map<int, gsl_matrix_float*>&inputSrc,
					     std::map<int, gsl_matrix_float*>&outputSrc,
					     FitsManager& mgr)
{
  unsigned long	nbExt = mgr.getNExtend();
  unsigned long	extCnt = 1;
  unsigned long	nbBlocs = 0;
  unsigned	row;
  unsigned	col;
  gsl_matrix_float	*inputPart;
  gsl_matrix_float	*outputPart;
  gsl_vector_float	*inputVect = gsl_vector_float_alloc(input->size2);
  
  if (inputSrc.size() == 0)
    return setError("DefectTrainer::createVectors", NO_EXT);
  else if (input == NULL || output == NULL)
    return setError("DefectTrainer::createVectors", NOT_ALLOC_MAT);
    
  while (extCnt <= nbExt)
    {
      inputPart = inputSrc[extCnt];
      outputPart = outputSrc[extCnt];
      
      if (inputPart != NULL && outputPart != NULL)
	{
	  while (nbBlocs < input->size1 / nbExt * extCnt)
	    {
	      row = (unsigned) ((inputPart->size1 - _blocSize) * (rand() / (RAND_MAX + 1.0)));
	      col = (unsigned) ((inputPart->size2 - _blocSize) * (rand() / (RAND_MAX + 1.0)));

	      // code de porc (_PCA->) a revoir
	      _PCA->getVectorfromBloc(inputVect, inputPart, row, col);

	      if (priorBloc(inputVect, extCnt))
		{
		  gsl_matrix_float_set_row(input, nbBlocs, inputVect);
		  gsl_matrix_float_set(output, nbBlocs, 0, gsl_matrix_float_get(outputPart, row + _blocSize / 2, col + _blocSize / 2));
		  ++nbBlocs;
		}
	    }
	  if (_printRebuilt == false)
	    unreadImageExtension(extCnt);
	  unreadMaskExtension(extCnt);
	}
      else
	{
	  gsl_vector_float_free(inputVect);
	  return setError("DefectTrainer::createVectors", NO_EXT);
	}
      ++extCnt;
    }
  gsl_vector_float_free(inputVect);
  return setError("DefectTrainer::createInputVector", NO_ERROR);
}

///////////////////////////////////////////////////////////////////////////////
const bool	DefectTrainer::priorBloc(const gsl_vector_float * const vect, const unsigned long extCnt)
{
  double	val;

  val = 0;
  for (unsigned cnt = 0; cnt < vect->size; ++cnt)
    {
      val += gsl_vector_float_get(vect, cnt);
    }
  val /= vect->size;
  double max = _imagesInfo[extCnt].max;
  if (((val >= _minPrior * max) && 
       (val <= _maxPrior * max)) || 
      (rand() / (RAND_MAX + 1.0) <= _nonPrior))
    return true;
  return false;
}

///////////////////////////////////////////////////////////////////////////////
void		DefectTrainer::createKLBase(gsl_matrix_float *inputV)
{
  struct stat	stat;
  unsigned	cnt = 0;

  gsl_matrix_float *PCAInputV = gsl_matrix_float_alloc(inputV->size1, inputV->size2);
  if (PCAInputV == NULL)
    setError("DefectTrainer::createKLBase", MALLOC);
  
  while (cnt < PCAInputV->size1)
    {
      unsigned row = (unsigned) ((inputV->size1) * (rand() / (RAND_MAX + 1.0)));
      
      for (unsigned cnt2 = 0; cnt2 < inputV->size2; ++cnt2)
	gsl_matrix_float_set(PCAInputV, cnt, cnt2, gsl_matrix_float_get(inputV, row, cnt2));

      ++cnt;
    }

  if (lstat(_evecFile.c_str(), &stat) == 0)
    {
      FILE		*f;
      unsigned long	size1;
      unsigned long	size2;
      
      f = fopen(_evecFile.c_str(), "r");
      fread(&size1, sizeof(unsigned long), 1, f);
      fread(&size2, sizeof(unsigned long), 1, f);
      if ((size1 != inputV->size2) || (size2 != inputV->size2))
	std::cout << "KL-base dimensions are bad (code setError())" << std::endl;
      gsl_matrix_float *evec = gsl_matrix_float_alloc(size1, size2);
      gsl_matrix_float_fread(f, evec);
      fclose(f);
      printMessage("KL-base loaded");
      _PCA->computePCA(PCAInputV, _nbPC, evec);
    }
  else
    {
      _PCA->computePCA(PCAInputV, _nbPC);
    }

  _PCA->computePCA(inputV, _nbPC);
  gsl_matrix_float_free(PCAInputV);
}

///////////////////////////////////////////////////////////////////////////////
const errorCodes	DefectTrainer::createNetwork(gsl_matrix_float *inputV, 
						     gsl_matrix_float *outputV)
{
  char		msg[128];
  if (createVectors(inputV, outputV, _images, _masks, _imageMgr) != NO_ERROR)
    return getError();
  sprintf(msg, "%d input vectors %d elements long created", 
	  inputV->size1, inputV->size2);
  printMessage(msg);
  createKLBase(inputV);

  unsigned size = _PCA->getPrincipalComponents()->size2;
  _l0 = create_network2(size, lround(size / 2.0), 1, SIGMOID_0_1);
  init_network(_l0);
  _l0->energy_treshold = _energyTreshold;
  _l0->max_iter = _maxIter;
  printMessage("Network created");
  gsl_rprop(_l0, outputV->size1,
	    _PCA->getPrincipalComponents(),
	    outputV);

  double	mean = 0;
  for (unsigned cnt = 0; cnt < _PCA->getPrincipalComponents()->size1; ++cnt)
    {
      feed_forward(_l0, (double *)gsl_matrix_float_const_ptr
		   (_PCA->getPrincipalComponents(), 
		    cnt , 0));
      mean += pow(gsl_matrix_float_get(outputV, cnt, 0) - _l0->output_values[0], 2);
    }
  mean /= _PCA->getPrincipalComponents()->size1;
  sprintf(msg, "quadratic error %f", mean);
  printMessage(msg);
  printMessage("Network trained using rprop algorithm");
  save_network_config(_l0, _nnFile.c_str());
  sprintf(msg, "Network saved to %s", _nnFile.c_str());
  printMessage(msg);
  free_network(_l0);
  return setError("DefectTrainer::createFirstNetwork", NO_ERROR);
}


///////////////////////////////////////////////////////////////////////////////
const errorCodes	DefectTrainer::train()
{
  gsl_matrix_float		*inputV;
  gsl_matrix_float		*outputV;

  inputV = gsl_matrix_float_alloc(_nbVectors, _blocSize * _blocSize);
  outputV = gsl_matrix_float_alloc(_nbVectors, 1);

  if (createNetwork(inputV, outputV) != NO_ERROR)
    return getError();

  gsl_matrix_float_free(inputV);
  gsl_matrix_float_free(outputV);

  return setError("DefectTrainer::createNetworks", NO_ERROR);
}
